"""
Label guesser guesses label based on document text and bayes filters.
It adds the guessed labels on new documents (in other words, freshly
added documents).

It stores data in 2 ways:
- sklearn pickling of TfidVectorizer
- sqlite database
"""
import collections
import gc
import logging
import os
import pickle
import sqlite3
import threading
import time

import numpy
import scipy.sparse
import sklearn.feature_extraction.text
import sklearn.naive_bayes

import openpaperwork_core
import openpaperwork_core.promise

from .... import (_, sync)


LOGGER = logging.getLogger(__name__)
ID = "label_guesser"
CREATE_TABLES = [
    # we could use the labels file instead, but some people access their work
    # directory through very slow Internet connections, so we better
    # keep a copy in the DB.
    (
        "CREATE TABLE IF NOT EXISTS labels ("
        " doc_id TEXT NOT NULL,"
        " label TEXT NOT NULL,"
        " PRIMARY KEY (doc_id, label)"
        ")"
    ),
    (  # see sklearn.feature_extraction.text.CountVectorizer.vocabulary
        "CREATE TABLE IF NOT EXISTS vocabulary ("
        " word TEXT NOT NULL,"
        " feature INTEGER NOT NULL,"
        " PRIMARY KEY (feature)"
        " UNIQUE (word)"
        ")"
    ),
    (
        "CREATE TABLE IF NOT EXISTS features ("
        " doc_id TEXT NOT NULL,"
        " vector NUMPY_ARRAY NOT NULL,"
        " PRIMARY KEY (doc_id)"
        ")"
    )
]


class SqliteNumpyArrayHandler(object):
    @staticmethod
    def _to_sqlite(np_array):
        return np_array.tobytes()

    @staticmethod
    def _from_sqlite(raw):
        return numpy.frombuffer(raw)

    @classmethod
    def register(cls):
        sqlite3.register_adapter(numpy.array, cls._to_sqlite)
        sqlite3.register_converter("NUMPY_ARRAY", cls._from_sqlite)


class PluginConfig(object):
    SETTINGS = {  # settings --> default value
        'batch_size': 200,
        # Limit the number of words used for performance reasons
        'max_words': 15000,
        # Limit the number of documents used for performance reasons
        # backlog is the number of required documents with and without each
        # label
        'max_doc_backlog': 100,
        'max_time': 10,  # seconds

        # if the document contains too few words, the classifiers tend to
        # put every possible labels on it. --> we ignore documents with too
        # few words / features.
        'min_features': 10,
    }

    def __init__(self, core):
        self.core = core

    def register(self):
        class Setting(object):
            def __init__(s, setting, default_val):
                s.setting = setting
                s.default_val = default_val

            def register(s):
                setting = self.core.call_success(
                    "config_build_simple", "label_guessing",
                    s.setting, lambda: s.default_val
                )
                self.core.call_all(
                    "config_register", "label_guessing_" + s.setting, setting
                )

        for (setting, default_val) in self.SETTINGS.items():
            Setting(setting, default_val).register()

    def get(self, key):
        return self.core.call_success("config_get", "label_guessing_" + key)


class UpdatableVectorizer(object):
    """
    Vectorizer that we can update over time with new features
    (word identifiers). Store the word->features in the SQLite database.
    We only add words to the database, we never remove them, otherwise we
    wouldn't be able to read the feature vectors correctly anymore.
    """
    def __init__(self, core, db_cursor):
        self.core = core
        self.db_cursor = db_cursor

        vocabulary = self.db_cursor.execute(
            "SELECT word, feature FROM vocabulary"
        )
        vocabulary = {k.strip(): v for (k, v) in vocabulary}
        if "" in vocabulary:
            vocabulary.pop("")

        self.updatable_vocabulary = vocabulary
        self.last_feature_id = max(vocabulary.values(), default=-1)

    def partial_fit_transform(self, corpus):
        # A bit hackish: We just need the analyzer, so instantiating a full
        # TfidVectorizer() is probably overkill, but meh.
        tokenizer = sklearn.feature_extraction.text.TfidfVectorizer(
        ).build_analyzer()
        LOGGER.info(
            "Vocabulary contains %d words before fitting", self.last_feature_id
        )
        for doc_txt in corpus:
            for word in tokenizer(doc_txt):
                if word in self.updatable_vocabulary:
                    continue
                self.last_feature_id += 1
                self.db_cursor.execute(
                    "INSERT INTO vocabulary (word, feature)"
                    " SELECT ?, ?"
                    "  WHERE NOT EXISTS("
                    "   SELECT 1 FROM vocabulary WHERE word = ?"
                    "  )",
                    (word.strip(), self.last_feature_id, word)
                )
                self.updatable_vocabulary[word.strip()] = self.last_feature_id
        LOGGER.info(
            "Vocabulary contains %d words after fitting", self.last_feature_id
        )

        return self.transform(corpus)

    def transform(self, corpus):
        # IMPORTANT: we must use use_idf=False here because we want the values
        # in each feature vector to be independant from other vectors.
        try:
            vectorizer = sklearn.feature_extraction.text.TfidfVectorizer(
                use_idf=False, vocabulary=self.updatable_vocabulary
            )
            features = vectorizer.fit_transform(corpus)
            LOGGER.info("%s features extracted", features.shape)
            return features
        except ValueError as exc:
            LOGGER.warning("Failed to extract features", exc_info=exc)
            return scipy.sparse.csr_matrix((0, 0))

    def _find_unused(self):
        doc_features = self.db_cursor.execute(
            "SELECT vector FROM features"
        )
        sum_features = None
        for (doc_vector,) in doc_features:
            if doc_vector is None:
                continue
            required_padding = (
                self.last_feature_id + 1 - doc_vector.shape[0]
            )
            if required_padding > 0:
                doc_vector = numpy.hstack([
                    doc_vector,
                    numpy.zeros((required_padding,))
                ])
            if sum_features is None:
                sum_features = doc_vector
            else:
                sum_features = sum_features + doc_vector

        if sum_features is None:
            return ([], 0)

        return (
            [f for (f, v) in enumerate(sum_features) if v == 0.0],
            len(sum_features)
        )

    def _get_all_doc_ids(self):
        doc_ids = self.db_cursor.execute("SELECT doc_id FROM features")
        doc_ids = [doc_id[0] for doc_id in doc_ids]
        return doc_ids

    def gc(self):
        """
        Drop features that are unused anymore.

        IMPORTANT: After this call, the vectorizer must not be used anymore
        (this method doesn't update the internal state of the vectorizer)
        """
        # we work in the main thread so we don't have to load all the feature
        # vectors all at once in memory (we just need their sum)
        LOGGER.info("Garbage collecting unused features ...")
        (to_drop, total) = self._find_unused()
        if len(to_drop) <= 0:
            LOGGER.info("No features to garbage collect (total=%d)", total)
            return
        LOGGER.info(
            "%d/%d features will be removed from the database",
            len(to_drop), total
        )

        doc_ids = self._get_all_doc_ids()

        # first we reduce the document feature vectors
        msg = _(
            "Label guesser: Garbage-collecting unused document features ..."
        )
        self.core.call_all("on_progress", "label_vector_gc", 0.0, msg)
        for (idx, doc_id) in enumerate(doc_ids):
            self.core.call_all(
                "on_progress", "label_vector_gc", idx / len(doc_ids), msg
            )
            doc_vector = self.db_cursor.execute(
                "SELECT vector FROM features WHERE doc_id = ? LIMIT 1",
                (doc_id,)
            )
            doc_vector = list(doc_vector)
            doc_vector = doc_vector[0][0]
            if doc_vector is None:
                # may happen according to bug report #478
                doc_vector = numpy.array([])

            to_drop_for_this_doc = [f for f in to_drop if f < len(doc_vector)]
            doc_vector = numpy.delete(doc_vector, to_drop_for_this_doc)
            self.db_cursor.execute(
                "UPDATE features SET vector = ? WHERE doc_id = ?",
                (doc_vector, doc_id)
            )
        self.core.call_all("on_progress", "label_vector_gc", 1.0)

        # then we reduce the vocabulary accordingly
        msg = _("Label guesser: Garbage-collecting unused words ...")
        self.core.call_all("on_progress", "label_vocabulary_gc", 0.0, msg)
        for (idx, f) in enumerate(to_drop):
            self.core.call_all(
                "on_progress", "label_vocabulary_gc", idx / len(to_drop), msg
            )
            self.db_cursor.execute(
                "DELETE FROM vocabulary WHERE feature = ?", (f,)
            )
            self.db_cursor.execute(
                "UPDATE vocabulary SET feature = feature - 1"
                " WHERE feature >= ?", (f,)
            )
        self.core.call_all("on_progress", "label_vocabulary_gc", 1.0, )

    def copy(self):
        r = UpdatableVectorizer(self.core, self.db_cursor)
        r.updatable_vocabulary = dict(self.updatable_vocabulary)
        r.last_feature_id = self.last_feature_id
        return r


class BatchIterator(object):
    def __init__(self, config, features, targets):
        self.features = features
        self.targets = targets
        self.b = 0
        self.batch_size = config.get("batch_size")

    def __iter__(self):
        self.b = 0
        return self

    def __next__(self):
        batch_corpus = self.features[self.b:self.b + self.batch_size]
        if len(batch_corpus) <= 0:
            raise StopIteration()
        batch_corpus = scipy.sparse.vstack(batch_corpus).toarray()
        batch_targets = self.targets[self.b:self.b + self.batch_size]
        self.b += self.batch_size
        return (batch_corpus, batch_targets)


class FeatureReductor(object):
    def __init__(self, to_drop):
        self.features_to_drop = to_drop

    def reduce_features(self, features):
        return numpy.delete(features, self.features_to_drop)


class DummyFeatureReductor(object):
    def reduce_features(self, features):
        return features


class Corpus(object):
    """
    Handles doc_ids and their associate feature vectors.
    Make sure we train the document in the best order possible, so even
    if the training is interrupted (time limit), we still have some training
    for most labels.
    """
    def __init__(self, config, doc_ids, features, targets):
        self.config = config
        self.doc_ids = doc_ids
        self.features = features
        self.targets = targets

    def standardize_feature_vectors(self, vectorizer):
        for idx in range(0, len(self.features)):
            doc_vector = self.features[idx]
            required_padding = (
                vectorizer.last_feature_id + 1 - doc_vector.shape[0]
            )
            assert required_padding >= 0
            if required_padding > 0:
                doc_vector = scipy.sparse.hstack([
                    scipy.sparse.csr_matrix(doc_vector),
                    numpy.zeros((1, required_padding))
                ])
            else:
                doc_vector = scipy.sparse.csr_matrix(doc_vector)
            self.features[idx] = doc_vector

    def reduce_corpus_words(self):
        """
        We may end up with a lot of different words (about 76000 in my case).
        But most of them are actually too rare to be useful and they use
        a lot memory and CPU time.
        """
        max_words = self.config.get("max_words")

        word_freq_sums = sum(self.features).toarray()[0]
        word_count = word_freq_sums.shape[0]
        LOGGER.info("Total word count before reduction: %d", word_count)
        if word_count <= max_words:
            LOGGER.info("No reduction to do")
            return DummyFeatureReductor()

        threshold = sorted(word_freq_sums, reverse=True)[max_words]
        LOGGER.info("Word frequency threshold: %f", threshold)

        features_to_drop = []
        for (idx, freq) in enumerate(word_freq_sums):
            if freq > threshold:
                continue
            features_to_drop.append(idx)

        reductor = FeatureReductor(features_to_drop)

        for idx in range(0, len(self.features)):
            self.features[idx] = scipy.sparse.csr_matrix(
                reductor.reduce_features(
                    self.features[idx].toarray()[0]
                )
            )

        LOGGER.info(
            "Total word count after reduction: %d",
            word_count - len(reductor.features_to_drop)
        )
        return reductor

    def get_doc_count(self):
        return len(self.features)

    def get_labels(self):
        return self.targets.keys()

    def get_batches(self, label):
        return BatchIterator(self.config, self.features, self.targets[label])

    @staticmethod
    def _add_doc_ids(max_doc_backlog, doc_weights, doc_ids):
        # Assumes doc_ids are in reverse order (most recent first)
        # Also assumes most recent documents are the most useful
        assert len(doc_ids) <= max_doc_backlog
        weigth = max_doc_backlog + 1
        for doc_id in doc_ids:
            doc_weights[doc_id] += weigth
            weigth -= 1

    @staticmethod
    def load(config, cursor):
        start = time.time()

        # doc_id --> weigth
        doc_weights = collections.defaultdict(lambda: 0)

        all_labels = cursor.execute("SELECT DISTINCT label FROM labels")
        all_labels = {label[0] for label in all_labels}
        all_docs = cursor.execute(
            "SELECT doc_id FROM features ORDER BY doc_id DESC"
        )
        all_docs = [doc[0] for doc in all_docs]

        max_doc_backlog = config.get("max_doc_backlog")

        for label in all_labels:
            ds = cursor.execute(
                "SELECT doc_id FROM labels WHERE label = ?"
                " ORDER BY doc_id DESC LIMIT {}".format(max_doc_backlog),
                (label,)
            )
            Corpus._add_doc_ids(
                max_doc_backlog, doc_weights, [d[0] for d in ds]
            )

        # label --> number of doc without this label
        no_label_counts = {label: 0 for label in all_labels}
        no_label_docids = {label: [] for label in all_labels}
        for doc_id in all_docs:
            if len(no_label_counts) <= 0:
                break
            doc_labels = cursor.execute(
                "SELECT label FROM labels WHERE doc_id = ?", (doc_id,)
            )
            for label in list(no_label_counts.keys()):
                if label in doc_labels:
                    continue
                no_label_docids[label].append(doc_id)
                no_label_counts[label] += 1
                if no_label_counts[label] >= max_doc_backlog:
                    no_label_counts.pop(label)
        for doc_ids in no_label_docids.values():
            doc_ids.sort(reverse=True)
            Corpus._add_doc_ids(max_doc_backlog, doc_weights, doc_ids)

        LOGGER.info(
            "Loading features of %d documents for %d labels",
            len(doc_weights), len(all_labels)
        )

        all_features = {}
        for doc_id in doc_weights.keys():
            vectors = cursor.execute(
                "SELECT vector FROM features WHERE doc_id = ? LIMIT 1",
                (doc_id,)
            )
            for vector in vectors:
                vector = vector[0]
                if vector is None:
                    continue
                all_features[doc_id] = vector

        all_features = [
            [weight, doc_id, all_features[doc_id]]
            for (doc_id, weight) in doc_weights.items()
            if doc_id in all_features
        ]
        all_features.sort(reverse=True)
        doc_ids = [
            doc_id
            for (weight, doc_id, features) in all_features
        ]
        features = [
            features
            for (weight, doc_id, features) in all_features
        ]

        # Load labels
        targets = collections.defaultdict(list)
        for (idx, doc_id) in enumerate(doc_ids):
            doc_labels = cursor.execute(
                "SELECT label FROM labels WHERE doc_id = ?",
                (doc_id,)
            )
            doc_labels = [label[0] for label in doc_labels]
            for label in all_labels:
                present = label in doc_labels
                targets[label].append(1 if present else 0)

        corpus = Corpus(
            config=config,
            doc_ids=doc_ids,
            features=features,
            targets=targets
        )

        stop = time.time()

        LOGGER.info(
            "Took %dms to load features of %d documents",
            int((stop - start) * 1000), len(doc_ids)
        )

        return corpus


class LabelGuesserTransaction(sync.BaseTransaction):
    def __init__(self, plugin, guess_labels=False, total_expected=-1):
        super().__init__(plugin.core, total_expected)
        self.priority = plugin.PRIORITY

        self.plugin = plugin
        self.guess_labels = guess_labels

        LOGGER.info("Transaction start: Guessing labels: %s", guess_labels)

        self.cursor = None
        self.vectorizer = None
        self.lock_acquired = False

        self.nb_changes = 0
        self.need_gc = False

    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.cancel()

    def _lazyinit_transaction(self):
        if self.cursor is not None:
            return

        # prevent reload of classifiers during the transaction
        assert not self.lock_acquired
        self.plugin.classifiers_cond.acquire()
        self.lock_acquired = True

        if not self.plugin.classifiers_loaded and self.guess_labels:
            # classifiers haven't been loaded at all yet. Now
            # looks like a good time for it (end of initial sync)
            self.plugin.reload_label_guessers()
        if self.plugin.classifiers is None and self.guess_labels:
            # Before starting the transaction, we wait for the classifiers
            # to be loaded, because we may need them
            # (see add_doc() -> _set_guessed_labels())
            self.plugin.classifiers_cond.wait()

        self.cursor = self.plugin.core.call_success(
            "sqlite_open",
            self.plugin.sql_url,
            detect_types=sqlite3.PARSE_DECLTYPES,
        )
        self.cursor.execute("BEGIN TRANSACTION")
        self.vectorizer = UpdatableVectorizer(self.core, self.cursor)

    def add_doc(self, doc_id):
        if self.guess_labels:
            # we have a higher priority than index plugins, so it is a good
            # time to update the document labels

            doc_url = self.core.call_success("doc_id_to_url", doc_id)

            has_labels = self.core.call_success(
                "doc_has_labels_by_url", doc_url
            )
            if has_labels is not None:
                LOGGER.info(
                    "Document %s already has labels. Won't guess labels",
                    doc_url
                )
                return

            self._lazyinit_transaction()
            self.plugin._set_guessed_labels(doc_url)

        else:
            self._lazyinit_transaction()

        self.notify_progress(
            ID, _("Label guesser: added document %s") % doc_id
        )
        self._upd_doc(doc_id)
        self.nb_changes += 1
        super().add_doc(doc_id)

    def upd_doc(self, doc_id):
        self._lazyinit_transaction()

        self.notify_progress(
            ID, _("Label guesser: updated document %s") % doc_id
        )
        self._upd_doc(doc_id)
        self.nb_changes += 1
        super().upd_doc(doc_id)

    def _del_doc(self, doc_id):
        self.cursor.execute("DELETE FROM labels WHERE doc_id = ?", (doc_id,))
        self.cursor.execute("DELETE FROM features WHERE doc_id = ?", (doc_id,))

    def del_doc(self, doc_id):
        self._lazyinit_transaction()

        self.notify_progress(
            ID,
            _("Label guesser: deleted document %s") % doc_id
        )
        self.nb_changes += 1
        self._del_doc(doc_id)
        LOGGER.info(
            "Document %s has been deleted."
            " Feature garbage-collecting will be run",
            doc_id
        )
        self.need_gc = True
        super().del_doc(doc_id)

    def _upd_doc(self, doc_id):
        self._del_doc(doc_id)

        doc_url = self.core.call_success("doc_id_to_url", doc_id)

        doc_labels = set()
        if doc_url is not None:
            self.core.call_all("doc_get_labels_by_url", doc_labels, doc_url)
        doc_labels = {label[0] for label in doc_labels}
        for doc_label in doc_labels:
            self.cursor.execute(
                "INSERT INTO labels (doc_id, label)"
                " VALUES (?, ?)",
                (doc_id, doc_label)
            )

        doc_txt = []
        self.core.call_all("doc_get_text_by_url", doc_txt, doc_url)
        doc_txt = "\n\n".join(doc_txt).strip()

        vector = self.vectorizer.partial_fit_transform([doc_txt])
        if vector.shape[0] <= 0 or vector.shape[1] <= 0:
            vector = numpy.array([])
        else:
            vector = vector[0].toarray()

        self.cursor.execute(
            "INSERT INTO features (doc_id, vector) VALUES (?, ?)",
            (doc_id, vector)
        )

    def cancel(self):
        try:
            self.core.call_success(
                "mainloop_schedule", self.core.call_all,
                "on_label_guesser_canceled"
            )
            if self.cursor is not None:
                self.cursor.execute("ROLLBACK")
                self.core.call_success(
                    "sqlite_close", self.cursor, optimize=False
                )
                self.cursor = None
            self.notify_done(ID)

            if not self.plugin.classifiers_loaded:
                # classifiers haven't been loaded at all yet. Now
                # looks like a good time for it (end of initial sync)
                self.plugin.reload_label_guessers()
        finally:
            if self.lock_acquired:
                self.plugin.classifiers_cond.release()
                self.lock_acquired = False

    def commit(self):
        try:
            LOGGER.info("Committing")

            if self.nb_changes > 0 and os.path.exists(self.plugin.model_path):
                os.unlink(self.plugin.model_path)

            self.core.call_success(
                "mainloop_schedule", self.core.call_all,
                "on_label_guesser_commit_start"
            )
            if self.nb_changes <= 0:
                assert self.cursor is None
                self.core.call_success(
                    "mainloop_schedule", self.core.call_all,
                    'on_label_guesser_commit_end'
                )
                LOGGER.info("Nothing to do. Training left unchanged.")
                if self.plugin.classifiers is None:
                    # classifiers haven't been loaded at all yet. Now
                    # looks like a good time for it (end of initial sync)
                    self.plugin.reload_label_guessers()
                return

            if self.need_gc:
                self.vectorizer.gc()
                self.vectorizer = None

            self.notify_progress(
                ID, _("Commiting changes for label guessing ...")
            )

            self.cursor.execute("COMMIT")
            self.core.call_success("sqlite_close", self.cursor)

            self.core.call_success(
                "mainloop_schedule", self.core.call_all,
                'on_label_guesser_commit_end'
            )
            LOGGER.info("Label guessing updated")

            self.plugin.reload_label_guessers()
        finally:
            self.notify_done(ID)
            if self.lock_acquired:
                self.plugin.classifiers_cond.release()
                self.lock_acquired = False


class Plugin(openpaperwork_core.PluginBase):
    PRIORITY = 100

    def __init__(self):
        self.bayes = None
        self.config = None

        self.sql = None

        # If we load the classifiers, we keep the vectorizer that goes with
        # them, because the vectorizer must return vectors that have the size
        # that the classifiers expect. If we would train the vectorizer with
        # new documents, the vector sizes could increase
        # Since loading the classifiers is a fairly long operations, we only
        # do it when we actually need them.
        self.word_reductor = None
        self.classifiers = None
        self.vectorizer = None

        # indicates whether the classifiers have been / are
        # been loaded
        self.classifiers_loaded = False

        # Do NOT use an RLock() here: The transaction locks this mutex
        # until commit() (or cancel()) is called. However, add_doc(),
        # del_doc(), upd_doc() and commit() may be called from different
        # threads.
        self.classifiers_cond = threading.Condition(threading.Lock())

        self.bayes_dir = None
        self.sql_url = None
        # We pickle the trained model. But pickle and Gio don't work well
        # together --> therefore we work directly with a path here instead of
        # an URL and with `os` instead `self.core.call_xxx("fs_xxx")`
        self.model_path = None

        SqliteNumpyArrayHandler.register()

    def get_interfaces(self):
        return [
            'sync',
        ]

    def get_deps(self):
        return [
            {
                'interface': 'config',
                'defaults': ['openpaperwork_core.config'],
            },
            {
                'interface': 'data_dir_handler',
                'defaults': ['paperwork_backend.datadirhandler'],
            },
            {
                'interface': 'data_versioning',
                'defaults': ['openpaperwork_core.data_versioning'],
            },
            {
                'interface': 'doc_labels',
                'defaults': ['paperwork_backend.model.labels'],
            },
            {
                'interface': 'doc_tracking',
                'defaults': ['paperwork_backend.doctracker'],
            },
            {
                'interface': 'document_storage',
                'defaults': ['paperwork_backend.model.workdir'],
            },
            {
                'interface': 'fs',
                'defaults': ['openpaperwork_gtk.fs.gio'],
            },
            {
                'interface': 'mainloop',
                'defaults': ['openpaperwork_gtk.mainloop.glib'],
            },
            {
                'interface': 'sqlite',
                'defaults': ['openpaperwork_core.sqlite'],
            },
            {
                'interface': 'transaction_manager',
                'defaults': ['paperwork_backend.sync'],
            },
            {
                'interface': 'work_queue',
                'defaults': ['openpaperwork_core.work_queue.default'],
            },
        ]

    def init(self, core):
        super().init(core)

        self.config = PluginConfig(core)
        self.config.register()

        self.core.call_all(
            "work_queue_create", "label_sklearn", stop_on_quit=True
        )

        self._init()

    def _init(self):
        data_dir = self.core.call_success(
            "data_dir_handler_get_individual_data_dir"
        )

        if self.bayes_dir is None:  # may be set by tests
            self.bayes_dir = self.core.call_success(
                "fs_join", data_dir, "bayes"
            )

        self.core.call_success("fs_mkdir_p", self.bayes_dir)

        self.sql_url = self.core.call_success(
            "fs_join", self.bayes_dir, 'label_guesser.db'
        )
        self.model_path = self.core.call_success(
            "fs_unsafe", self.core.call_success(
                "fs_join", self.bayes_dir, "model.pickle"
            )
        )
        LOGGER.info("Opening guesswork.label.sklearn database")
        self.sql = self.core.call_one(
            "sqlite_execute",
            self.core.call_success,
            "sqlite_open",
            self.sql_url,
            detect_types=sqlite3.PARSE_DECLTYPES,
            timeout=60
        )
        for query in CREATE_TABLES:
            self.sql.execute(query)

        self.core.call_all(
            "doc_tracker_register",
            "label_guesser",
            self._build_transaction
        )

    def _build_transaction(self, sync, total_expected):
        return LabelGuesserTransaction(
            self, guess_labels=not sync, total_expected=total_expected
        )

    def on_data_dir_changed(self):
        self.on_quit()
        self._init()

    def on_quit(self):
        LOGGER.info("Closing guesswork.label.sklearn database")
        self.core.call_one(
            "sqlite_execute",
            self.core.call_success,
            "sqlite_close", self.sql
        )

    def reload_label_guessers(self):
        self.classifiers = None
        self.classifiers_loaded = True
        promise = openpaperwork_core.promise.ThreadedPromise(
            self.core, self._reload_label_guessers
        )
        self.core.call_all("work_queue_cancel_all", "label_sklearn")
        self.core.call_success(
            "work_queue_add_promise", "label_sklearn", promise
        )

    def _reload_label_guessers(self):
        with self.classifiers_cond:
            try:
                cursor = self.core.call_one(
                    "sqlite_open",
                    self.sql_url,
                    detect_types=sqlite3.PARSE_DECLTYPES,
                )
                try:
                    cursor.execute("BEGIN TRANSACTION")
                    self.vectorizer = UpdatableVectorizer(self.core, cursor)
                    (
                        self.word_reductor, self.classifiers
                    ) = self._load_classifiers(
                        self.model_path, cursor, self.vectorizer
                    )
                finally:
                    cursor.execute("ROLLBACK")
                    self.core.call_one("sqlite_close", cursor, optimize=False)
            finally:
                self.classifiers_cond.notify_all()

    def _load_classifiers(self, model_path, cursor, vectorizer):
        if os.path.exists(model_path):
            with open(model_path, "rb") as fd:
                LOGGER.info("Pre-trained model available. Using it")
                (reductor, classifiers) = pickle.load(fd)
            return (reductor, classifiers)

        LOGGER.info("Pre-trained model not available. Re-training one")

        # Jflesch> This is a very memory-intensive process. The Glib may try
        # to allocate memory before the GC runs and AFAIK the Glib
        # is not aware of Python GC, and so won't trigger it if required
        # (instead it will abort).
        # --> free as much memory as possible now
        # (Remember that there may be 32bits version of Paperwork out there)
        gc.collect()

        msg = _("Label guessing: Training ...")

        try:
            self.core.call_all("on_progress", "label_classifiers", 0.0, msg)

            corpus = Corpus.load(self.config, cursor)

            LOGGER.info("Training classifiers ...")
            start = time.time()

            if corpus.get_doc_count() <= 1:
                return (DummyFeatureReductor(), {})

            corpus.standardize_feature_vectors(vectorizer)
            # no need to train on all the words. Only the most used words
            reductor = corpus.reduce_corpus_words()

            # Jflesch> This is a very memory-intensive process. The Glib may
            # try to allocate memory before the GC runs and AFAIK the Glib
            # is not aware of Python GC, and so won't trigger it if required
            # (instead it will abort).
            # --> free as much memory as possible now
            gc.collect()

            classifiers = collections.defaultdict(
                sklearn.naive_bayes.GaussianNB
            )

            batch_iterators = [
                (label, corpus.get_batches(label))
                for label in corpus.get_labels()
            ]
            done = 0
            total = len(batch_iterators) * corpus.get_doc_count()

            loop_nb = 0
            timeout = False

            max_time = self.config.get("max_time")
            fit_start = time.time()
            try:
                while not timeout:
                    for (label, batch_iterator) in batch_iterators:
                        now = time.time()
                        if loop_nb > 0 and now - fit_start > max_time:
                            timeout = True
                            LOGGER.warning(
                                "Training is taking too long (%dms > %dms)."
                                " Interrupting",
                                (now - fit_start) * 1000, max_time * 1000
                            )
                            break

                        (batch_corpus, batch_targets) = next(batch_iterator)
                        self.core.call_all(
                            "on_progress", "label_classifiers", done / total,
                            _("Label guessing: Training ...")
                        )
                        classifiers[label].partial_fit(
                            batch_corpus, batch_targets,
                            classes=[0, 1]
                        )
                        done += len(batch_corpus)

                    loop_nb += 1
            except StopIteration:
                pass

            stop = time.time()
            LOGGER.info(
                "Training took %dms (Fitting: %dms) ;"
                " Training completed at %d%%",
                (stop - start) * 1000,
                (stop - fit_start) * 1000,
                done * 100 / total
            )

            # Jflesch> This is a very memory-intensive process. The Glib may
            # try to allocate memory before the GC runs and AFAIK the Glib
            # is not aware of Python GC, and so won't trigger it if required
            # (instead it will abort).
            # --> free as much memory as possible now
            gc.collect()

            with open(model_path, "wb") as fd:
                pickle.dump((reductor, classifiers), fd)
            return (reductor, classifiers)
        finally:
            self.core.call_all("on_progress", "label_classifiers", 1.0, msg)

    def _guess(self, vectorizer, reductor, classifiers, doc_url):
        LOGGER.info("Guessing labels on %s", doc_url)
        doc_txt = []
        self.core.call_all("doc_get_text_by_url", doc_txt, doc_url)
        doc_txt = "\n\n".join(doc_txt).strip()
        if doc_txt == u"":
            LOGGER.info("Can't guess. Text is empty")
            return

        vector = vectorizer.transform([doc_txt])
        vector = vector.toarray()
        if len(vector) <= 0:
            LOGGER.info(
                "Failed to vectorize text (text length=%d)",
                len(doc_txt)
            )
            return
        vector = vector[0]
        vector = reductor.reduce_features(vector)

        min_features = self.config.get("min_features")
        nb_features = 0
        for f in vector:
            if f > 0:
                nb_features += 1

        if nb_features < min_features:
            LOGGER.warning(
                "Document doesn't contain enough different words"
                " (%d ; min required is %d). Labels won't be guessed",
                nb_features, min_features
            )
            return

        LOGGER.info("Documents contains %d features", nb_features)

        for (label, classifier) in classifiers.items():
            predicted = classifier.predict([vector])[0]
            if predicted:
                yield label

    def _set_guessed_labels(self, doc_url):
        # self.classifiers_cond must locked
        assert self.classifiers is not None
        labels = self._guess(
            self.vectorizer, self.word_reductor, self.classifiers, doc_url
        )
        labels = list(labels)
        for label in labels:
            self.core.call_success("doc_add_label_by_url", doc_url, label)

    def tests_cleanup(self):
        self.core.call_success("sqlite_close", self.sql)



class AbstractPageEditorUI(object):
    CAPABILITY_SHOW_FRAME = (1 << 0)
    CAPABILITIES = 0

    def can(self, capability):
        return bool(self.CAPABILITIES & capability)

    def set_modifier_state(self, modifier_id, enabled):
        """
        Indicates if an modifier must be shown as enabled or disabled.
        """
        pass

    def show_preview(self, img):
        """
        Indicates the image to display, with all the changes applied from the
        modifiers already applied on it.

        Only called when the image changed.
        """
        return

    def show_frame_selector(self):
        """
        Tells the UI that it must let the user select a frame on the image
        (image provided by `show_image`). It also specify the current frame
        to display to the user.

        Called every time the image changes if we want the frame to be shown.

        Frame can be obtained by calling PageEditor.frame.get() and
        can be updated with PageEditor.frame.set().
        """
        return

    def hide_frame_selector(self):
        """
        Tells the UI to not let the user select a frame anymore.
        """
        return

    def on_edit_end(self, doc_url, page_idx):
        """
        Called when the user modifications have been applied or cancelled.
        """
        pass



import logging
import os

from .. import PluginBase


LOGGER = logging.getLogger(__name__)


class Plugin(PluginBase):
    PRIORITY = 100

    def get_interfaces(self):
        return ['external_apps']

    def get_deps(self):
        return [
            {
                'interface': 'fs',
                'defaults': ['openpaperwork_core.fs.python'],
            },
        ]

    def external_app_open_file(self, file_url):
        if not hasattr(os, 'startfile'):
            return None
        # os.startfile() is Windows-only.
        LOGGER.info("Opening %s with os.startfile()", file_url)
        assert file_url.startswith("file://")
        file_path = self.core.call_success("fs_safe", file_url)
        os.startfile(file_path)
        return True

    def external_app_open_folder(self, folder_url):
        return self.external_app_open_file(folder_url)

    def external_app_can_send_as_attachment(self) -> None:
        return None

    def external_app_send_as_attachment(self, uri) -> None:
        return None



import logging


LOGGER = logging.getLogger(__name__)


class Task(object):
    def __init__(self, core, func, args, kwargs):
        self.core = core
        self.func = func
        self.args = args
        self.kwargs = kwargs
        # The mainloop can't track other threads, but if there is
        # a graceful shutdown waiting, we don't want it to stop the main
        # loop before our thread is done.
        # --> increment mainloop ref counter before
        core.call_all("mainloop_ref", self)

    def __str__(self):
        return "Task<{}>({}, {})".format(self.func, self.args, self.kwargs)

    def __repr__(self):
        return str(self)

    def do(self):
        try:
            self.func(*self.args, **self.kwargs)
        except Exception as exc:
            LOGGER.error(
                "==== UNCAUGHT EXCEPTION IN THREAD ===", exc_info=exc
            )
        finally:
            self.core.call_all("mainloop_unref", self)



import logging

import openpaperwork_core


LOGGER = logging.getLogger(__name__)


class _LogHandler(logging.Handler):
    def __init__(self, plugin):
        super().__init__()
        self.plugin = plugin
        self.log_level = logging.DEBUG
        self.formatter = logging.Formatter(Plugin.LOG_FORMAT)

    def emit(self, record):
        if self.plugin.console is None:
            return
        if record.levelno < self.log_level:
            return
        line = self.formatter.format(record)
        self.plugin.console.print(line)


class Plugin(openpaperwork_core.PluginBase):
    LOG_FORMAT = '[%(levelname)-6s] [%(name)-30s] %(message)s'
    LOG_LEVELS = {
        'none': logging.CRITICAL,
        'critical': logging.CRITICAL,
        'error': logging.ERROR,
        'warn': logging.WARN,
        'warning': logging.WARNING,
        'info': logging.INFO,
        'debug': logging.DEBUG,
    }

    def __init__(self):
        super().__init__()
        self.console = None

    def get_interfaces(self):
        return [
            'logs',
            'uncaught_exception_listener',
        ]

    def get_deps(self):
        return [
            {
                'interface': 'uncaught_exception',
                'defaults': ['openpaperwork_core.uncaught_exception'],
            },
        ]

    def cmd_set_console(self, console):
        self.console = console

    def init_logs(self, app_name, default_log_level):
        self.log_handler = _LogHandler(self)
        logging.getLogger().addHandler(self.log_handler)

        lvl = self.LOG_LEVELS[default_log_level]
        self.log_handler.log_level = lvl
        if lvl > logging.INFO:
            # Never disable info level ; it may be used by other plugins
            lvl = logging.INFO
        logging.getLogger().setLevel(lvl)

    def on_uncaught_exception(self, exc_info):
        LOGGER.error("=== UNCAUGHT EXCEPTION ===", exc_info=exc_info)
        LOGGER.error("==========================")



import os
import tempfile
import time
import unittest

import openpaperwork_core


class AbstractTestSafe(unittest.TestCase):
    def get_plugin_name(self):
        """
        must be subclassed
        """
        assert False

    def setUp(self):
        self.core = openpaperwork_core.Core(auto_load_dependencies=True)
        self.core.load(self.get_plugin_name())
        self.core.init()

    @unittest.skipUnless(os.name == 'posix', reason="Linux only")
    def test_linux(self):
        v = self.core.call_one('fs_safe', '/home/flesch jerome')
        self.assertEqual(v, "file:///home/flesch%20jerome")

        v = self.core.call_one('fs_safe', 'file:///home/flesch%20jerome')
        self.assertEqual(v, "file:///home/flesch%20jerome")

    @unittest.skipUnless(os.name == 'nt', reason="Windows only")
    def test_windows(self):
        v = self.core.call_one('fs_safe', 'c:\\Users\\flesch jerome')
        self.assertEqual(v, "file:///c:/Users/flesch%20jerome")

        v = self.core.call_one('fs_safe', '\\\\someserver\\someshare')
        self.assertEqual(v, "\\\\someserver\\someshare")


class AbstractTestUnsafe(unittest.TestCase):
    def get_plugin_name(self):
        """
        must be subclassed
        """
        assert False

    def setUp(self):
        self.core = openpaperwork_core.Core(auto_load_dependencies=True)
        self.core.load(self.get_plugin_name())
        self.core.init()

    @unittest.skipUnless(os.name == 'posix', reason="Linux only")
    def test_linux(self):
        v = self.core.call_one('fs_unsafe', 'file:///home/flesch%20jerome')
        self.assertEqual(v, "/home/flesch jerome")

        v = self.core.call_one('fs_unsafe', 'file:///home/flesch jerome')
        self.assertEqual(v, "/home/flesch jerome")

        v = self.core.call_one('fs_unsafe', '/home/flesch jerome')
        self.assertEqual(v, "/home/flesch jerome")

    @unittest.skipUnless(os.name == 'nt', reason="Windows only")
    def test_windows(self):
        v = self.core.call_success(
            'fs_unsafe', 'file://c:\\Users\\flesch%20jerome'
        )
        self.assertEqual(v, "c:\\Users\\flesch jerome")

        v = self.core.call_one('fs_safe', '\\\\someserver\\someshare')
        self.assertEqual(v, "\\\\someserver\\someshare")


class AbstractTestOpen(unittest.TestCase):
    def get_plugin_name(self):
        """
        must be subclassed
        """
        assert False

    def setUp(self):
        self.core = openpaperwork_core.Core(auto_load_dependencies=True)
        self.core.load(self.get_plugin_name())
        self.core.init()

    def test_read_binary(self):
        file_name = None
        with tempfile.NamedTemporaryFile(mode='wb', delete=False) as fd:
            file_name = fd.name
            fd.write(b"content")
        try:
            safe_file_name = self.core.call_one('fs_safe', file_name)
            with self.core.call_one(
                        'fs_open', safe_file_name, mode='rb'
                    ) as fd:
                content = fd.read()
                self.assertEqual(content, b"content")
        finally:
            os.unlink(file_name)

    def test_write_binary(self):
        file_name = None
        with tempfile.NamedTemporaryFile(mode='wb', delete=False) as fd:
            file_name = fd.name
        try:
            safe_file_name = self.core.call_one('fs_safe', file_name)
            with self.core.call_one(
                        'fs_open', safe_file_name, mode='wb'
                    ) as fd:
                fd.write(b"content\n")
            with open(file_name, 'rb') as fd:
                content = fd.read()
                self.assertEqual(content, b"content\n")
        finally:
            os.unlink(file_name)

    def test_read(self):
        file_name = None
        with tempfile.NamedTemporaryFile(mode='w', delete=False) as fd:
            file_name = fd.name
            fd.write("content")
        try:
            safe_file_name = self.core.call_one('fs_safe', file_name)
            with self.core.call_one('fs_open', safe_file_name, mode='r') as fd:
                content = fd.read()
                self.assertEqual(content, "content")
        finally:
            os.unlink(file_name)

    def test_write(self):
        file_name = None
        with tempfile.NamedTemporaryFile(mode='w', delete=False) as fd:
            file_name = fd.name
        try:
            safe_file_name = self.core.call_one('fs_safe', file_name)
            with self.core.call_one('fs_open', safe_file_name, mode='w') as fd:
                fd.write("content\n")
            with open(file_name, 'r') as fd:
                content = fd.read()
                self.assertEqual(content, "content\n")
        finally:
            os.unlink(file_name)


class AbstractTestExists(unittest.TestCase):
    def get_plugin_name(self):
        """
        must be subclassed
        """
        assert False

    def setUp(self):
        self.core = openpaperwork_core.Core(auto_load_dependencies=True)
        self.core.load(self.get_plugin_name())
        self.core.init()

    def test_not_exist(self):
        self.assertEqual(self.core.call_one('fs_exists', "/whatever"), None)

    def test_exists(self):
        file_name = None
        with tempfile.NamedTemporaryFile(mode='w', delete=False) as fd:
            file_name = fd.name
            fd.write("content")
        try:
            safe_file_name = self.core.call_one('fs_safe', file_name)
            self.assertTrue(self.core.call_one('fs_exists', safe_file_name))
        finally:
            os.unlink(file_name)


class AbstractTestListDir(unittest.TestCase):
    def get_plugin_name(self):
        """
        must be subclassed
        """
        assert False

    def setUp(self):
        self.core = openpaperwork_core.Core(auto_load_dependencies=True)
        self.core.load(self.get_plugin_name())
        self.core.init()

        self.dirname = tempfile.mkdtemp()
        self.uri_dirname = self.core.call_one('fs_safe', self.dirname)
        with open(os.path.join(self.dirname, 'test1.txt'), 'w') as fd:
            fd.write('test1')
        with open(os.path.join(self.dirname, 'test2.txt'), 'w') as fd:
            fd.write('test2')
        os.mkdir(os.path.join(self.dirname, 'test3'))
        with open(os.path.join(self.dirname, 'test3', 'test4.txt'), 'w') as fd:
            fd.write('test4')

    def tearDown(self):
        # check fs_rm_rf at the same time
        self.core.call_one("fs_rm_rf", self.uri_dirname, trash=False)
        self.assertFalse(os.path.exists(self.dirname))

    def test_listdir(self):
        files = list(self.core.call_one('fs_listdir', self.uri_dirname))
        files.sort()

        self.assertEqual(files, [
            self.uri_dirname + '/test1.txt',
            self.uri_dirname + '/test2.txt',
            self.uri_dirname + '/test3'
        ])

    def test_recurse(self):
        files = list(self.core.call_one('fs_recurse', self.uri_dirname))
        files.sort()

        self.assertEqual(files, [
            self.uri_dirname + '/test1.txt',
            self.uri_dirname + '/test2.txt',
            self.uri_dirname + '/test3/test4.txt'
        ])


class AbstractTestRename(unittest.TestCase):
    def get_plugin_name(self):
        """
        must be subclassed
        """
        assert False

    def setUp(self):
        self.core = openpaperwork_core.Core(auto_load_dependencies=True)
        self.core.load(self.get_plugin_name())
        self.core.init()

    def test_rename(self):
        src_file_name = None
        dst_file_name = None
        with tempfile.NamedTemporaryFile(mode='w', delete=False) as fd:
            src_file_name = fd.name
            fd.write("content")
        with tempfile.NamedTemporaryFile(mode='w', delete=False) as fd:
            dst_file_name = fd.name
            fd.write("content")
        os.unlink(dst_file_name)

        try:
            safe_src_file_name = self.core.call_one('fs_safe', src_file_name)
            safe_dst_file_name = self.core.call_one('fs_safe', dst_file_name)
            self.assertTrue(os.path.exists(src_file_name))
            self.assertFalse(os.path.exists(dst_file_name))

            self.core.call_one(
                'fs_rename', safe_src_file_name, safe_dst_file_name
            )
            self.assertFalse(os.path.exists(src_file_name))
            self.assertTrue(os.path.exists(dst_file_name))
        finally:
            if os.path.exists(src_file_name):
                os.unlink(src_file_name)
            if os.path.exists(dst_file_name):
                os.unlink(dst_file_name)


class AbstractTestUnlink(unittest.TestCase):
    def get_plugin_name(self):
        """
        must be subclassed
        """
        assert False

    def setUp(self):
        self.core = openpaperwork_core.Core(auto_load_dependencies=True)
        self.core.load(self.get_plugin_name())
        self.core.init()

    def test_unlink(self):
        file_name = None
        with tempfile.NamedTemporaryFile(mode='w', delete=False) as fd:
            file_name = fd.name
            fd.write("content")

        try:
            safe_file_name = self.core.call_one('fs_safe', file_name)
            self.assertTrue(os.path.exists(file_name))

            self.core.call_one('fs_unlink', safe_file_name, trash=False)
            self.assertFalse(os.path.exists(file_name))
        finally:
            if os.path.exists(file_name):
                os.unlink(file_name)


class AbstractTestGetMtime(unittest.TestCase):
    def get_plugin_name(self):
        """
        must be subclassed
        """
        assert False

    def setUp(self):
        self.core = openpaperwork_core.Core(auto_load_dependencies=True)
        self.core.load(self.get_plugin_name())
        self.core.init()

    def test_get_mtime(self):
        now = time.time()

        file_name = None
        with tempfile.NamedTemporaryFile(mode='w', delete=False) as fd:
            file_name = fd.name
            fd.write("content")

        try:
            safe_file_name = self.core.call_one('fs_safe', file_name)
            mtime = self.core.call_one('fs_get_mtime', safe_file_name)
            self.assertTrue(int(now) <= mtime)
            self.assertTrue(mtime <= now + 2)

            time.sleep(2)  # :-(

            with open(file_name, 'a') as fd:
                fd.write("content")
            new_mtime = self.core.call_one('fs_get_mtime', safe_file_name)
            self.assertTrue(int(now) <= new_mtime)
            self.assertTrue(mtime < new_mtime)
        finally:
            os.unlink(file_name)


class AbstractTestGetsize(unittest.TestCase):
    def get_plugin_name(self):
        """
        must be subclassed
        """
        assert False

    def setUp(self):
        self.core = openpaperwork_core.Core(auto_load_dependencies=True)
        self.core.load(self.get_plugin_name())
        self.core.init()

    def test_getsize(self):
        file_name = None
        with tempfile.NamedTemporaryFile(mode='w', delete=False) as fd:
            file_name = fd.name
            fd.write("content")  # 7 bytes

        try:
            safe_file_name = self.core.call_one('fs_safe', file_name)

            s = self.core.call_one('fs_getsize', safe_file_name)
            self.assertEqual(s, 7)
        finally:
            os.unlink(file_name)


class AbstractTestIsdir(unittest.TestCase):
    def get_plugin_name(self):
        """
        must be subclassed
        """
        assert False

    def setUp(self):
        self.core = openpaperwork_core.Core(auto_load_dependencies=True)
        self.core.load(self.get_plugin_name())
        self.core.init()

    def test_isdir_true(self):
        dirname = tempfile.mkdtemp()
        uri_dirname = self.core.call_one('fs_safe', dirname)

        try:
            self.assertTrue(self.core.call_one("fs_isdir", uri_dirname))
        finally:
            # check fs_rm_rf at the same time
            self.core.call_one("fs_rm_rf", uri_dirname, trash=False)
            self.assertFalse(os.path.exists(dirname))

    def test_isdir_false(self):
        file_name = None
        with tempfile.NamedTemporaryFile(mode='w', delete=False) as fd:
            file_name = fd.name
            fd.write("content")

        try:
            safe_file_name = self.core.call_one('fs_safe', file_name)

            self.assertFalse(self.core.call_one('fs_isdir', safe_file_name))
        finally:
            os.unlink(file_name)


class AbstractTestCopy(unittest.TestCase):
    def get_plugin_name(self):
        """
        must be subclassed
        """
        assert False

    def setUp(self):
        self.core = openpaperwork_core.Core(auto_load_dependencies=True)
        self.core.load(self.get_plugin_name())
        self.core.init()

    def test_copy(self):
        src_file_name = None
        dst_file_name = None
        with tempfile.NamedTemporaryFile(mode='w', delete=False) as fd:
            src_file_name = fd.name
            fd.write("content")
        with tempfile.NamedTemporaryFile(mode='w', delete=False) as fd:
            dst_file_name = fd.name
            fd.write("xxxxxxx")
        os.unlink(dst_file_name)

        try:
            safe_src_file_name = self.core.call_one('fs_safe', src_file_name)
            safe_dst_file_name = self.core.call_one('fs_safe', dst_file_name)
            self.assertTrue(os.path.exists(src_file_name))
            self.assertFalse(os.path.exists(dst_file_name))

            self.core.call_one(
                'fs_copy', safe_src_file_name, safe_dst_file_name
            )
            self.assertTrue(os.path.exists(src_file_name))
            self.assertTrue(os.path.exists(dst_file_name))

            with open(dst_file_name, 'r') as fd:
                content = fd.read()
                self.assertEqual(content, 'content')
        finally:
            if os.path.exists(src_file_name):
                os.unlink(src_file_name)
            if os.path.exists(dst_file_name):
                os.unlink(dst_file_name)


class AbstractTestMkdirP(unittest.TestCase):
    def get_plugin_name(self):
        """
        must be subclassed
        """
        assert False

    def setUp(self):
        self.core = openpaperwork_core.Core(auto_load_dependencies=True)
        self.core.load(self.get_plugin_name())
        self.core.init()

    def test_mkdir_p(self):
        dirname = tempfile.mkdtemp()
        uri_dirname = self.core.call_one('fs_safe', dirname)
        try:
            new_dirs = uri_dirname + "/testA/testB/testC"
            self.core.call_one("fs_mkdir_p", new_dirs)
            self.assertTrue(os.path.exists(os.path.join(
                dirname, "testA", "testB", "testC"
            )))
        finally:
            # check fs_rm_rf at the same time
            self.core.call_one("fs_rm_rf", uri_dirname, trash=False)
            self.assertFalse(os.path.exists(dirname))


class AbstractTestBasename(unittest.TestCase):
    def get_plugin_name(self):
        """
        must be subclassed
        """
        assert False

    def setUp(self):
        self.core = openpaperwork_core.Core(auto_load_dependencies=True)
        self.core.load(self.get_plugin_name())
        self.core.init()

    def test_basename(self):
        out = self.core.call_success("fs_basename", "file:///a/b/c.txt")
        self.assertEqual(out, "c.txt")
        out = self.core.call_success("fs_basename", "file:///c.txt")
        self.assertEqual(out, "c.txt")
        out = self.core.call_success("fs_basename", "memory://camion.txt")
        self.assertEqual(out, "camion.txt")


class AbstractTestTemp(unittest.TestCase):
    def get_plugin_name(self):
        """
        must be subclassed
        """
        assert False

    def setUp(self):
        self.core = openpaperwork_core.Core(auto_load_dependencies=True)
        self.core.load(self.get_plugin_name())
        self.core.init()

    def test_mktemp(self):
        (tmp_url, tmp_fd) = self.core.call_success(
            "fs_mktemp", prefix="test", suffix=".txt", mode="w"
        )
        with tmp_fd:
            tmp_fd.write("TEST\n")

        self.assertNotEqual(
            self.core.call_success("fs_exists", tmp_url), None
        )

        self.core.call_all("on_quit")

        self.assertEqual(
            self.core.call_success("fs_exists", tmp_url), None
        )



import threading

from . import Task
from .. import PluginBase


class Plugin(PluginBase):
    """
    Simply create a thread for each task to run. Less efficient than a thread
    pool, but may be useful for testing or debugging.
    """
    def get_interfaces(self):
        return ['thread']

    def get_deps(self):
        return [
            {
                'interface': 'mainloop',
                'defaults': ['openpaperwork_core.mainloop.asyncio'],
            },
        ]

    def thread_start(self, func, *args, **kwargs):
        task = Task(self.core, func, args, kwargs)
        thread = threading.Thread(target=task.do)
        thread.daemon = True
        thread.start()
        return True



"""
When drag'n'dropping, enables the dropping-in-a-document part.
"""

import logging

import openpaperwork_core
import openpaperwork_core.promise

from . import BaseDocViewController


LOGGER = logging.getLogger(__name__)


class DropController(BaseDocViewController):
    LINE_BORDERS = 10
    LINE_WIDTH = 3
    LINE_COLOR = (0.0, 0.8, 1.0, 1.0)

    def __init__(self, core, plugin):
        super().__init__(plugin)
        self.core = core
        self.closest = None
        self.beforeafter = -1  # 0 for before the widget, 1 for after

    @staticmethod
    def _compute_squared_distance(rect, x, y):
        pts = (
            (rect.x, rect.y),
            (rect.x + rect.width, rect.y),
            (rect.x, rect.y + rect.height),
            (rect.x + rect.width, rect.y + rect.height),
        )
        dists = {
            ((abs(pt[0] - x) ** 2) + (abs(pt[1] - y) ** 2))
            for pt in pts
        }
        return min(dists)

    def _compute_closest(self, x, y):
        dists = {
            (
                self._compute_squared_distance(widget.get_allocation(), x, y),
                widget
            )
            for widget in self.plugin.page_layout.get_children()
        }
        try:
            return min(dists)[1]
        except ValueError:
            # empty list of widgets
            return None

    @staticmethod
    def _compute_beforeafter(rect, x, y):
        center = rect.x + (rect.width / 2)
        return 0 if x < center else 1

    def on_drag_motion(self, drag_context, x, y, time):
        super().on_drag_motion(drag_context, x, y, time)

        self.closest = self._compute_closest(x, y)
        if self.closest is None:
            self.beforeafter = -1
            return

        self.beforeafter = self._compute_beforeafter(
            self.closest.get_allocation(), x, y
        )

        self.plugin.page_layout.queue_draw()

    def on_drag_leave(self, drag_context, time):
        super().on_drag_leave(drag_context, time)
        self.closest = None
        self.plugin.page_layout.queue_draw()

    def on_draw(self, cairo_ctx):
        super().on_draw(cairo_ctx)

        if self.closest is None:
            return

        allocation = self.closest.get_allocation()
        height = allocation.height
        x = allocation.x
        x -= 2 * self.LINE_BORDERS
        x += (self.beforeafter * (allocation.width + 2 * self.LINE_BORDERS))
        x = max(0, x)
        y = allocation.y

        cairo_ctx.save()
        try:
            cairo_ctx.set_source_rgba(*self.LINE_COLOR)
            cairo_ctx.set_line_width(self.LINE_WIDTH)

            cairo_ctx.move_to(x + self.LINE_BORDERS, y)
            cairo_ctx.line_to(x + self.LINE_BORDERS, y + height)
            cairo_ctx.stroke()

            cairo_ctx.move_to(x, y + self.LINE_BORDERS)
            cairo_ctx.line_to(
                x + (2 * self.LINE_BORDERS),
                y + self.LINE_BORDERS
            )
            cairo_ctx.stroke()

            cairo_ctx.move_to(x, y + height - self.LINE_BORDERS)
            cairo_ctx.line_to(
                x + (2 * self.LINE_BORDERS),
                y + height - self.LINE_BORDERS
            )
            cairo_ctx.stroke()
        finally:
            cairo_ctx.restore()

    def drag_and_drop_get_destination(self, x, y):
        super().drag_and_drop_get_destination(x, y)
        closest = self._compute_closest(x, y)
        if closest is None:
            dst_page_idx = 0
        else:
            page = self.plugin.widget_to_page[closest]
            beforeafter = self._compute_beforeafter(
                closest.get_allocation(), x, y
            )
            dst_page_idx = page.page_idx + beforeafter
        return (*self.plugin.active_doc, dst_page_idx)


class Plugin(openpaperwork_core.PluginBase):
    def __init__(self):
        super().__init__()

    def get_interfaces(self):
        return [
            'chkdeps',
            'gtk_docview_controller',
        ]

    def get_deps(self):
        return [
            {
                'interface': 'document_storage',
                'defaults': ['paperwork_backend.model.workdir'],
            },
            {
                'interface': 'gtk_docview',
                'defaults': ['paperwork_gtk.mainwindow.docview'],
            },
            {
                'interface': 'gtk_drag_and_drop',
                'defaults': ['paperwork_gtk.gesture.drag_and_drop'],
            },
        ]

    def on_gtk_docview_init(self, docview):
        self.core.call_all("drag_and_drop_page_enable", docview.page_layout)

    def gtk_docview_get_controllers(self, out: dict, docview):
        out['drop'] = DropController(self.core, docview)



"""
Crop scanned images based on a predefined area.
"""

import logging

import openpaperwork_core

from ... import (_, sync)
from . import ID


LOGGER = logging.getLogger(__name__)


class CalibrationTransaction(sync.BaseTransaction):
    def __init__(self, plugin, sync, total_expected=-1):
        super().__init__(plugin.core, total_expected)

        self.priority = plugin.PRIORITY

        self.plugin = plugin
        self.sync = sync

        # For each document, we need to track on which pages we have already
        # guessed the page borders and on which page we didn't yet.
        # We use the same ID for all the cropping plugins so we never crop
        # twice the same page.
        self.page_tracker = self.core.call_success("page_tracker_get", ID)

    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.cancel()

    def _crop_page(self, doc_id, doc_url, page_idx, page_nb, total_pages):
        paper_size = self.core.call_success(
            "page_get_paper_size_by_url", doc_url, page_idx
        )
        if paper_size is not None:
            # We only want to crop scanned pages.
            self.notify_progress(
                ID,
                _("Document {doc_id} p{page_idx} already cropped").format(
                    doc_id=doc_id, page_idx=(page_idx + 1)
                ),
                page_nb=page_nb, total_pages=total_pages
            )
            return

        if self.core.call_success(
                    "page_has_text_by_url", doc_url, page_idx
                ):
            self.notify_progress(
                ID,
                _(
                    "Document {doc_id} p{page_idx} has already some text."
                    " Not cropping."
                ).format(doc_id=doc_id, page_idx=(page_idx + 1)),
                page_nb=page_nb, total_pages=total_pages
            )
            return

        self.notify_progress(
            ID,
            _(
                "Using calibration to crop page borders of"
                " document {doc_id} p{page_idx}"
            ).format(doc_id=doc_id, page_idx=(page_idx + 1)),
            page_nb=page_nb, total_pages=total_pages
        )
        self.plugin.crop_page_borders_by_url(doc_url, page_idx)

    def _crop_new_pages(self, doc_id):
        doc_url = self.core.call_success("doc_id_to_url", doc_id)

        modified_pages = list(self.page_tracker.find_changes(doc_id, doc_url))

        for (page_nb, (change, page_idx)) in enumerate(modified_pages):
            # Guess page borders on new pages, but only if we are
            # not currently synchronizing with the work directory
            # (when syncing we don't modify the documents, ever)
            if not self.sync and change == 'new':
                self._crop_page(
                    doc_id, doc_url, page_idx, page_nb, len(modified_pages)
                )
            self.page_tracker.ack_page(doc_id, doc_url, page_idx)

    def add_doc(self, doc_id):
        self._crop_new_pages(doc_id)
        super().add_doc(doc_id)

    def upd_doc(self, doc_id):
        self._crop_new_pages(doc_id)
        super().upd_doc(doc_id)

    def del_doc(self, doc_id):
        self.page_tracker.delete_doc(doc_id)
        super().del_doc(doc_id)

    def cancel(self):
        self.page_tracker.cancel()
        self.notify_done(ID)

    def commit(self):
        self.page_tracker.commit()
        self.notify_done(ID)


class Plugin(openpaperwork_core.PluginBase):
    PRIORITY = 4000

    def get_interfaces(self):
        return [
            "cropping",
            "scanner_calibration",
            "syncable",  # actually satisfied by the plugin 'doctracker'
        ]

    def get_deps(self):
        return [
            {
                'interface': 'config',
                'defaults': ['openpaperwork_core.config'],
            },
            {
                'interface': 'doc_tracking',
                'defaults': ['paperwork_backend.doctracker'],
            },
            {
                'interface': 'page_tracking',
                'defaults': ['paperwork_backend.pagetracker'],
            },
            {
                'interface': 'pillow',
                'defaults': [
                    'openpaperwork_core.pillow.img',
                    'paperwork_backend.pillow.pdf',
                ]
            }
        ]

    def init(self, core):
        super().init(core)
        self.core.call_all(
            "config_register", "scanner_calibration",
            self.core.call_success(
                "config_build_simple", "scanner", "calibration",
                lambda: None
            )
        )
        self.core.call_all(
            "doc_tracker_register", ID,
            lambda sync, total_expected=-1: CalibrationTransaction(
                self, sync, total_expected
            )
        )

    def crop_page_borders_by_url(self, doc_url, page_idx):
        frame = self.core.call_success(
            "config_get", "scanner_calibration"
        )
        if frame is None:
            LOGGER.warning(
                "No calibration found. Cannot crop page %s p%d",
                doc_url, page_idx
            )
            return None

        LOGGER.info(
            "Cropping page %d of %s (calibration=%s)",
            page_idx, doc_url, frame
        )

        doc_id = self.core.call_success("doc_url_to_id", doc_url)

        if doc_id is not None:
            self.core.call_one(
                "mainloop_schedule", self.core.call_all,
                "on_page_modification_start", doc_id, page_idx
            )
        try:
            page_img_url = self.core.call_success(
                "page_get_img_url", doc_url, page_idx
            )

            img = self.core.call_success("url_to_pillow", page_img_url)

            LOGGER.info(
                "Cropping page %d of %s at %s", page_idx, doc_url, frame
            )
            # make sure we don't extend the image
            frame = (
                max(0, frame[0]),
                max(0, frame[1]),
                min(img.size[0], frame[2]),
                min(img.size[1], frame[3]),
            )
            img = img.crop(frame)

            page_img_url = self.core.call_success(
                "page_get_img_url", doc_url, page_idx, write=True
            )
            self.core.call_success("pillow_to_url", img, page_img_url)
        finally:
            if doc_id is not None:
                self.core.call_one(
                    "mainloop_schedule", self.core.call_all,
                    "on_page_modification_end", doc_id, page_idx
                )
        return frame



import logging
import os
import shutil
import tempfile

from . import PluginBase


LOGGER = logging.getLogger(__name__)


class Plugin(PluginBase):
    PRIORITY = 75

    def __init__(self):
        super().__init__()
        self.tmp_files = set()

        self.flatpak = False
        self.tmp_dir = None

    def get_interfaces(self):
        return [
            'flatpak',
            'stats',
        ]

    def get_deps(self):
        return [
            {
                'interface': 'data_versioning',
                'defaults': ['openpaperwork_core.data_versioning'],
            },
            {
                'interface': 'fs',
                'defaults': ['openpaperwork_core.fs.python'],
            },
            {
                'interface': 'paths',
                'defaults': ['openpaperwork_core.paths.xdg'],
            },
        ]

    def init(self, core):
        super().init(core)
        self.flatpak = (os.name == 'posix') and bool(self.core.call_success(
            "fs_isdir", "file:///app"
        ))
        LOGGER.info("Flatpak environment: %s", self.flatpak)
        self.tmp_dir = self.core.call_success(
            "fs_join",
            self.core.call_success("paths_get_data_dir"), "tmp"
        )

    def is_in_flatpak(self):
        if self.flatpak:
            return True
        return None

    def fs_mktemp(self, prefix=None, suffix=None, mode='w+b', **kwargs):
        """
        Modifies slightly the behaviour of fs_mktemp():

        When making a bug report, we use temporary files. We need the user
        to be able to access those temporary files with external applications.

        With Flatpak, the easiest way for that is to place the temporary files
        somewhere in the user home directory.
        """
        if not self.flatpak:
            return None

        self.core.call_success("fs_mkdir_p", self.tmp_dir)
        tmp = tempfile.NamedTemporaryFile(
            prefix=prefix, suffix=suffix, delete=False, mode=mode,
            dir=self.core.call_success("fs_unsafe", self.tmp_dir)
        )
        LOGGER.info("Flatpak tmp file: %s --> %s", self.tmp_dir, tmp.name)
        self.tmp_files.add(tmp.name)
        return (self.core.call_success("fs_safe", tmp.name), tmp)

    def on_quit(self):
        if not self.flatpak:
            return
        if self.core.call_success("fs_exists", self.tmp_dir) is not None:
            shutil.rmtree(self.core.call_success("fs_unsafe", self.tmp_dir))

    def stats_get(self, out: dict):
        if not self.flatpak:
            return

        if 'os_name' in out:
            out['os_name'] += ' (flatpak)'
        else:
            out['os_name'] = 'GNU/Linux (flatpak)'



#    Paperwork - Using OCR to grep dead trees the easy way
#    Copyright (C) 2012-2019  Jerome Flesch
#
#    Paperwork is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Paperwork is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with Paperwork.  If not, see <http://www.gnu.org/licenses/>.
import openpaperwork_core
import openpaperwork_core.promise

from .. import _


DEFAULT_RESOLUTION = 300


class Plugin(openpaperwork_core.PluginBase):
    def __init__(self):
        super().__init__()
        self.out = []

    def get_interfaces(self):
        return ['shell']

    def get_deps(self):
        return [
            {
                "interface": "mainloop",
                "defaults": ['openpaperwork_gtk.mainloop.glib'],
            },
            {
                "interface": "scan2doc",
                "defaults": ["paperwork_backend.docscan.scan2doc"],
            },
        ]

    def cmd_complete_argparse(self, parser):
        scan_parser = parser.add_parser(
            'scan', help=_("Scan pages")
        )
        scan_parser.add_argument(
            "--doc_id", "-d", help=_(
                "Document to which the scanned pages must be added"
            )
        )
        scan_parser.add_argument("source_id")

    def cmd_run(self, console, args):
        if args.command != 'scan':
            return None

        doc_url = None
        if args.doc_id is not None:
            doc_url = self.core.call_success("doc_id_to_url", args.doc_id)

        promise = self.core.call_success(
            "scan2doc_promise",
            doc_id=args.doc_id, doc_url=doc_url,
            source_id=args.source_id
        )
        self.core.call_success("scan_schedule", promise)

        self.core.call_all("mainloop_quit_graceful")
        self.core.call_one("mainloop")

        return True



import logging

import openpaperwork_core
import openpaperwork_gtk.deps


try:
    import gi
    gi.require_version('Gtk', '3.0')
    from gi.repository import Gtk
    GTK_AVAILABLE = True
except (ImportError, ValueError):
    GTK_AVAILABLE = False


LOGGER = logging.getLogger(__name__)


class Plugin(openpaperwork_core.PluginBase):
    LAYOUTS = {
        'paged': {
            'icon': 'view-paged-symbolic',
        },
        'grid': {
            'icon': 'view-grid-symbolic',
        },
    }

    def __init__(self):
        super().__init__()
        self.widget_tree = None
        self.layout_icon = None
        self.layout_button = None
        self.notify = True

    def get_interfaces(self):
        return [
            'chkdeps',
            'gtk_layout_settings',
            'screenshot_provider',
        ]

    def get_deps(self):
        return [
            {
                'interface': 'gtk_docview',
                'defaults': ['paperwork_gtk.mainwindow.docview'],
            },
            {
                'interface': 'gtk_docview_pageinfo',
                'defaults': ['paperwork_gtk.mainwindow.docview.pageinfo'],
            },
            {
                'interface': 'gtk_resources',
                'defaults': ['openpaperwork_gtk.resources'],
            },
            {
                'interface': 'screenshot',
                'defaults': ['openpaperwork_gtk.screenshots'],
            },
        ]

    def chkdeps(self, out: dict):
        if not GTK_AVAILABLE:
            out['gtk'].update(openpaperwork_gtk.deps.GTK)

    def init(self, core):
        super().init(core)

        self.core.call_success(
            "gtk_load_css",
            "paperwork_gtk.mainwindow.docview.pageinfo", "layout_settings.css"
        )

        self.widget_tree = self.core.call_success(
            "gtk_load_widget_tree",
            "paperwork_gtk.mainwindow.docview.pageinfo",
            "layout_settings.glade"
        )
        if self.widget_tree is None:
            # init must still work so 'chkdeps' is still available
            LOGGER.error("Failed to load widget tree")
            return

        self.layout_icon = self.widget_tree.get_object("page_layout_icon")
        self.layout_button = self.widget_tree.get_object("page_layout")
        self.layout_button.connect("clicked", self.gtk_open_layout_settings)

        self.zoom = self.widget_tree.get_object("adjustment_zoom")
        self.zoom.connect("value-changed", self._on_zoom_changed)

        self.all_boxes = self.widget_tree.get_object("show_all_boxes")
        self.all_boxes.connect("notify::active", self._notify_all_boxes)

        self.layout_buttons = {
            self.widget_tree.get_object("layout_grid"): {
                "handler": None,
                "layout": "grid",
            },
            self.widget_tree.get_object("layout_paged"): {
                "handler": None,
                "layout": "paged",
            },
        }
        for button in self.layout_buttons.keys():
            self.layout_buttons[button]['handler'] = button.connect(
                "clicked", self._on_layout_change
            )

        self.core.call_success("page_info_add_left", self.layout_button)

        scroll = self.core.call_success("docview_get_scrollwindow")
        self.core.call_all("on_zoomable_widget_new", scroll, self.zoom)

    def on_layout_change(self, layout_name):
        if self.layout_icon is None:
            return

        icon = self.LAYOUTS[layout_name]['icon']
        # smallest icon size available
        self.layout_icon.set_from_icon_name(icon, Gtk.IconSize.SMALL_TOOLBAR)

    def _on_layout_change(self, widget):
        assert widget in self.layout_buttons
        layout = self.layout_buttons[widget]['layout']
        self.core.call_all("docview_set_layout", layout)

    def gtk_open_layout_settings(self, *args, **kwargs):
        menu = self.widget_tree.get_object("layout_settings")
        menu.set_relative_to(self.layout_button)
        menu.set_visible(True)

    def gtk_close_layout_settings(self, *args, **kwargs):
        menu = self.widget_tree.get_object("layout_settings")
        menu.set_visible(False)

    def _on_zoom_changed(self, _=None):
        new_value = self.zoom.get_value()
        if self.notify:
            self.core.call_all("docview_set_zoom", new_value)

    def _notify_all_boxes(self, *args, **kwargs):
        active = self.all_boxes.get_active()
        self.core.call_all("set_all_boxes_visible", active)

    def docview_set_zoom(self, zoom):
        current = self.zoom.get_value()
        if current != zoom:
            self.notify = False
            try:
                self.zoom.set_value(zoom)
            finally:
                self.notify = True

    def screenshot_snap_all_doc_widgets(self, out_dir):
        if self.widget_tree is None:
            return

        self.core.call_success(
            "screenshot_snap_widget",
            self.widget_tree.get_object("layout_settings"),
            self.core.call_success("fs_join", out_dir, "docview_layout.png"),
        )
        self.core.call_success(
            "screenshot_snap_widget",
            self.widget_tree.get_object("layout_paged"),
            self.core.call_success(
                "fs_join", out_dir, "docview_layout_paged.png"
            ),
            margins=(50, 50, 50, 50)
        )
        self.core.call_success(
            "screenshot_snap_widget",
            self.widget_tree.get_object("layout_grid"),
            self.core.call_success(
                "fs_join", out_dir, "docview_layout_grid.png"
            ),
            margins=(50, 50, 50, 50)
        )
        self.core.call_success(
            "screenshot_snap_widget",
            self.widget_tree.get_object("scale_zoom"),
            self.core.call_success(
                "fs_join", out_dir, "docview_layout_scale.png"
            ),
            margins=(50, 50, 50, 50)
        )
        self.core.call_success(
            "screenshot_snap_widget",
            self.widget_tree.get_object("show_all_boxes"),
            self.core.call_success(
                "fs_join", out_dir, "docview_layout_show_all_boxes.png"
            ),
            margins=(275, 50, 50, 50)
        )



import logging

try:
    import gi
    gi.require_version('Gdk', '3.0')
    from gi.repository import Gdk
    GTK_AVAILABLE = True
except (ImportError, ValueError):
    GTK_AVAILABLE = False

import openpaperwork_core
import openpaperwork_gtk.deps


LOGGER = logging.getLogger(__name__)


class Zoomer(object):
    ZOOM_INCREMENT = 0.02

    def __init__(self, zoomable, adjustment):
        self.adjustment = adjustment
        if zoomable.get_window() is not None:
            self._enable_events(zoomable)
        else:
            zoomable.connect("realize", self._enable_events)
        zoomable.connect("scroll-event", self._on_scroll)

    def _enable_events(self, zoomable):
        zoomable.add_events(Gdk.EventMask.SCROLL_MASK)
        zoomable.get_window().set_events(
            zoomable.get_window().get_events() | Gdk.EventMask.SCROLL_MASK
        )

    def _on_scroll(self, widget, event):
        if (event.state & Gdk.ModifierType.CONTROL_MASK):
            original = self.adjustment.get_value()
            delta = event.get_scroll_deltas()[2]
            if delta < 0:
                zoom = original + self.ZOOM_INCREMENT
            elif delta > 0:
                zoom = original - self.ZOOM_INCREMENT
            else:
                return False
            LOGGER.debug("Ctrl + Scrolling: zoom %f --> %f", original, zoom)
            self.adjustment.set_value(zoom)
            return True
        # don't know what to do, don't care. Let someone else take care of it
        return False


class Plugin(openpaperwork_core.PluginBase):
    def __init__(self):
        super().__init__()
        # XXX(Jflesch): Looks like a GObject Introspection bug: if the
        # Zoomer objects get garbage-collected, the signal isn't handled
        # anymore. --> We need to keep a reference to those objects as long
        # as we need them.
        self.refs = {}

    def get_interfaces(self):
        return [
            'chkdeps',
            'gtk_zoomable',
        ]

    def chkdeps(self, out: dict):
        if not GTK_AVAILABLE:
            out['gtk'].update(openpaperwork_gtk.deps.GTK)

    def on_zoomable_widget_new(self, zoomable, adjustment):
        """
        zoomable is the widget on which user can zoom.
        zoom gestures are not applied directly on the widget. They are
        applied on a GtkAdjustment.
        """
        zoomer = Zoomer(zoomable, adjustment)
        self.core.call_all("on_objref_track", zoomer)
        self.refs[zoomable] = zoomer

    def on_zoomable_widget_destroy(self, zoomable, adjustment):
        if zoomable in self.refs:
            self.refs.pop(zoomable)



import openpaperwork_core


class Plugin(openpaperwork_core.PluginBase):
    def get_interfaces(self):
        return []

    def get_deps(self):
        return [
            {
                'interface': 'gtk_progress_widget',
                'defaults': ['openpaperwork_gtk.widgets.progress'],
            },
            {
                'interface': 'gtk_docview',
                'defaults': ['paperwork_gtk.mainwindow.docview'],
            },
        ]

    def init(self, core):
        super().init(core)
        widget = self.core.call_success("gtk_progress_make_widget")
        if widget is None:
            # GTK is not available
            return
        header_bar = self.core.call_success("docview_get_headerbar")
        header_bar.pack_end(widget)



import logging

import pyocr
import pyocr.builders

import openpaperwork_core

from ... import (_, sync)


LOGGER = logging.getLogger(__name__)

ID = "ocr"


class OcrTransaction(sync.BaseTransaction):
    def __init__(self, plugin, sync, total_expected=-1):
        super().__init__(plugin.core, total_expected)

        self.priority = plugin.PRIORITY

        self.plugin = plugin
        self.sync = sync

        # for each document, we need to track which pages have already been
        # OCR-ed, which have been modified (cropping, rotation, ...)
        # and must be re-OCRed, and which have not been changed.
        self.page_tracker = self.core.call_success("page_tracker_get", ID)

    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.cancel()

    def _run_ocr_on_page(
            self, doc_id, doc_url, page_idx, page_nb, total_pages,
            wordless_only=False):
        if wordless_only:
            has_text = self.core.call_success(
                "page_has_text_by_url", doc_url, page_idx
            )
            if has_text:
                # there is already some text on this page
                self.notify_progress(
                    ID,
                    _(
                        "Document {doc_id} p{page_idx} has already some text."
                        " No OCR run"
                    ).format(doc_id=doc_id, page_idx=(page_idx + 1)),
                    page_nb=page_nb, total_pages=total_pages
                )
                return

        self.notify_progress(
            ID,
            _("Running OCR on document {doc_id} p{page_idx}").format(
                doc_id=doc_id, page_idx=(page_idx + 1)
            ),
            page_nb=page_nb, total_pages=total_pages
        )
        self.plugin.ocr_page_by_url(doc_url, page_idx)

    def _run_ocr_on_modified_pages(self, doc_id, wordless_only=False):
        doc_url = self.core.call_success("doc_id_to_url", doc_id)

        need_end_notification = False
        modified_pages = list(self.page_tracker.find_changes(doc_id, doc_url))
        for (page_nb, (change, page_idx)) in enumerate(modified_pages):
            # Run OCR on modified pages, but only if we are not synchronizing
            # with the work directory (--> if the user just added or modified
            # a document)
            if not self.sync and (change == 'new' or change == 'upd'):
                self._run_ocr_on_page(
                    doc_id, doc_url, page_idx, page_nb, len(modified_pages),
                    wordless_only
                )
                need_end_notification = True
            self.page_tracker.ack_page(doc_id, doc_url, page_idx)

        if need_end_notification:
            self.notify_progress(
                ID, _("Running OCR"),
                page_nb=len(modified_pages), total_pages=len(modified_pages)
            )

    def add_doc(self, doc_id):
        self._run_ocr_on_modified_pages(doc_id, wordless_only=True)
        super().add_doc(doc_id)

    def upd_doc(self, doc_id):
        self._run_ocr_on_modified_pages(doc_id, wordless_only=True)
        super().upd_doc(doc_id)

    def del_doc(self, doc_id):
        self.page_tracker.delete_doc(doc_id)
        super().del_doc(doc_id)

    def cancel(self):
        self.page_tracker.cancel()
        self.notify_done(ID)

    def commit(self):
        self.page_tracker.commit()
        self.notify_done(ID)


class Plugin(openpaperwork_core.PluginBase):
    PRIORITY = 1000

    def get_interfaces(self):
        return [
            "ocr",
            "syncable",  # actually satisfied by the plugin 'doctracker'
        ]

    def get_deps(self):
        return [
            {
                'interface': 'doc_tracking',
                'defaults': ['paperwork_backend.doctracker'],
            },
            {
                'interface': 'document_storage',
                'defaults': ['paperwork_backend.model.workdir'],
            },
            {
                'interface': 'ocr_settings',
                'defaults': ['paperwork_backend.pyocr'],
            },
            {
                'interface': 'page_boxes',
                'defaults': [
                    'paperwork_backend.model.hocr',
                    'paperwork_backend.model.pdf',
                ],
            },
            {
                'interface': 'page_tracking',
                'defaults': ['paperwork_backend.pagetracker'],
            },
            {
                'interface': 'pillow',
                'defaults': [
                    'openpaperwork_core.pillow.img',
                    'paperwork_backend.pillow.pdf',
                ],
            },
        ]

    def init(self, core):
        super().init(core)
        self.core.call_all(
            "doc_tracker_register", ID,
            lambda sync, total_expected=-1: OcrTransaction(
                self, sync, total_expected
            )
        )

    def ocr_page_by_url(self, doc_url, page_idx):
        if self.core.call_success("ocr_is_enabled") is None:
            LOGGER.info("OCR is disabled")
            return

        LOGGER.info("Running OCR on page %d of %s", page_idx, doc_url)

        doc_id = self.core.call_success("doc_url_to_id", doc_url)
        if doc_id is not None:
            self.core.call_one(
                "mainloop_schedule", self.core.call_all,
                "on_page_modification_start", doc_id, page_idx
            )
        try:
            page_img_url = self.core.call_success(
                "page_get_img_url", doc_url, page_idx
            )
            ocr_tool = pyocr.get_available_tools()[0]
            LOGGER.info(
                "Will use tool '%s' on %s p%d (%s)",
                ocr_tool.get_name(), doc_url, page_idx, page_img_url
            )

            ocr_langs = self.core.call_success("ocr_get_active_langs")

            img = self.core.call_success("url_to_pillow", page_img_url)

            boxes = ocr_tool.image_to_string(
                img, lang="+".join(ocr_langs),
                builder=pyocr.builders.LineBoxBuilder()
            )
            self.core.call_success(
                "page_set_boxes_by_url", doc_url, page_idx, boxes
            )
        except Exception as exc:
            LOGGER.error("OCR FAILED", exc_info=exc)
        finally:
            if doc_id is not None:
                self.core.call_one(
                    "mainloop_schedule", self.core.call_all,
                    "on_page_modification_end", doc_id, page_idx
                )

        return True



import logging

import openpaperwork_core
import paperwork_backend.sync


LOGGER = logging.getLogger(__name__)

DOC_ID = "help_intro"


class Plugin(openpaperwork_core.PluginBase):
    PRIORITY = -100000  # see storage_get_all_docs()

    def __init__(self):
        super().__init__()
        self.doc_url = None
        self.thumbnail_url = None
        self.opened = False
        self.deleted = False

    def get_interfaces(self):
        return [
            'doc_labels',
            'document_storage',
            'syncable',
        ]

    def get_deps(self):
        return [
            {
                'interface': 'help_documents',
                'defaults': ['paperwork_gtk.model.help'],
            },
        ]

    def init(self, core):
        super().init(core)

    def storage_get_all_docs(self, out: list, only_valid=True):
        if self.deleted or len(out) > 0:
            self.doc_url = None
            return

        self.doc_url = self.core.call_success("doc_id_to_url", DOC_ID)
        if (self.doc_url is None or
                not self.core.call_success("fs_exists", self.doc_url)):
            LOGGER.error(
                "Introduction document %s not found."
                " Was Paperwork packaged correctly ?",
                DOC_ID
            )
            self.doc_url = None
            return

        out.append((DOC_ID, self.doc_url))

        if not self.opened:
            self.opened = True
            self.core.call_success(
                "mainloop_schedule", self.core.call_all,
                "doc_open", DOC_ID, self.doc_url
            )
            self.core.call_success(
                "mainloop_schedule", self.core.call_all,
                "docview_set_layout", "paged"
            )

    def labels_get_all(self, out: set):
        if self.doc_url is not None:
            self.core.call_all("help_labels_get_all", out)

    def storage_delete_doc_id(self, doc_id):
        if doc_id != DOC_ID:
            return
        self.doc_url = None
        self.deleted = True

    def _fake_delete_doc(self):
        if self.doc_url is None:
            return

        all_docs = []
        self.core.call_all("storage_get_all_docs", all_docs)
        all_docs = [doc for doc in all_docs if doc[0] != DOC_ID]

        if len(all_docs) <= 0:
            return

        self.deleted = True
        # make sure the index is up-to-date
        self.core.call_success(
            "transaction_simple", [("del", DOC_ID),]
        )

    def doc_transaction_start(self, out: list, total_expected=-1):
        class FakeDeleteDocTransaction(paperwork_backend.sync.BaseTransaction):
            priority = -100000

            def commit(s):
                if self.doc_url is None:
                    return
                self._fake_delete_doc()

        out.append(FakeDeleteDocTransaction(self.core, total_expected))

    def sync(self, promises: list):
        if self.doc_url is None:
            return
        promises.append(openpaperwork_core.promise.Promise(
            self.core, self._fake_delete_doc
        ))



import openpaperwork_core


class Plugin(openpaperwork_core.PluginBase):
    def get_interfaces(self):
        return ['l10n_init']

    def get_deps(self):
        return [
            {
                'interface': 'l10n',
                'defaults': ['openpaperwork_core.l10n.python'],
            },
        ]

    def init(self, core):
        super().init(core)
        self.core.call_all(
            "l10n_load", "paperwork_backend.l10n", "paperwork_backend"
        )



import itertools
import logging

try:
    from gi.repository import Gio
    GLIB_AVAILABLE = True
except (ImportError, ValueError):
    GLIB_AVAILABLE = False

import openpaperwork_core
import openpaperwork_core.deps

from ... import _


LOGGER = logging.getLogger(__name__)
ACTION_NAME = "doc_change_labels"


class Plugin(openpaperwork_core.PluginBase):
    def __init__(self):
        super().__init__()
        self.menu_add = None
        self.menu_remove = None
        self.idx_generator = itertools.count()

    def get_interfaces(self):
        return [
            'chkdeps',
            'menu',
            'menu_docs',
            'menu_docs_properties',
        ]

    def get_deps(self):
        return [
            {
                'interface': 'action_docs_properties',
                'defaults': ['paperwork_gtk.actions.docs.properties'],
            },
            {
                'interface': 'docs_actions',
                'defaults': ['paperwork_gtk.mainwindow.doclist'],
            },
            {
                'interface': 'gtk_doclist',
                'defaults': ['paperwork_gtk.mainwindow.doclist'],
            },
        ]

    def chkdeps(self, out: dict):
        if not GLIB_AVAILABLE:
            out['glib'].update(openpaperwork_core.deps.GLIB)

    def on_doclist_initialized(self):
        item = Gio.MenuItem.new(_("Change labels"), "win." + ACTION_NAME)
        self.core.call_all("docs_menu_append_item", item)



from typing import Optional

import collections
import logging

try:
    from gi.repository import GLib  # noqa: F401
    GLIB_AVAILABLE = True
except (ImportError, ValueError):
    GLIB_AVAILABLE = False

try:
    import gi
    gi.require_version('Gtk', '3.0')
    from gi.repository import Gtk
    GTK_AVAILABLE = True
except (ImportError, ValueError):
    GTK_AVAILABLE = False

import PIL
import PIL.Image
import PIL.ImageDraw

import openpaperwork_core
import openpaperwork_core.deps
import openpaperwork_gtk.deps


LOGGER = logging.getLogger(__name__)


class Plugin(openpaperwork_core.PluginBase):
    def __init__(self):
        super().__init__()
        self.widget_tree = None
        self.core = None
        self.stacks = None
        self.components = collections.defaultdict(dict)
        self.navigation_stacks = {'left': [], 'right': []}

        self.defaults = collections.defaultdict(list)  # keep track of the prio
        self.mainwindow = None
        self._mainwindow_size = None

    def get_interfaces(self):
        return [
            'app_actions',
            'app_shortcuts',
            'chkdeps',
            'gtk_mainwindow',
            'screenshot_provider',
        ]

    def get_deps(self):
        return [
            {
                'interface': 'config',
                'defaults': ['openpaperwork_core.config'],
            },
            {
                'interface': 'fs',
                'defaults': ['openpaperwork_gtk.fs.gio']
            },
            {
                'interface': 'gtk_resources',
                'defaults': ['openpaperwork_gtk.resources'],
            },
            {
                'interface': 'gtk_init',
                'defaults': ['openpaperwork_gtk.gtk_init'],
            },
            {
                'interface': 'icon',
                'defaults': ['paperwork_gtk.icon'],
            },
            {
                'interface': 'l10n_init',
                'defaults': ['paperwork_gtk.l10n'],
            },
            {
                'interface': 'screenshot',
                'defaults': ['openpaperwork_gtk.screenshots'],
            },
        ]

    def init(self, core):
        super().init(core)

        self.core = core

        if not GTK_AVAILABLE:
            LOGGER.warning("not initializing main window without Gtk")
            return None

        self.core.call_success(
            "gtk_load_css",
            "paperwork_gtk.mainwindow", "global.css"
        )

        self.core.call_success(
            "gtk_load_css",
            "paperwork_gtk.mainwindow.window", "mainwindow.css"
        )

        self.widget_tree = self.core.call_success(
            "gtk_load_widget_tree",
            "paperwork_gtk.mainwindow.window", "mainwindow.glade"
        )
        if self.widget_tree is None:
            # init must still work so 'chkdeps' is still available
            LOGGER.error("Failed to load widget tree")
            return

        opt = self.core.call_success(
            "config_build_simple", "GUI", "main_window_size",
            lambda: (1024, 600)
        )
        self.core.call_all("config_register", "main_window_size", opt)
        main_win_size = self.core.call_success(
            "config_get", "main_window_size"
        )

        self.mainwindow = self.widget_tree.get_object("mainwindow")
        self.mainwindow.set_default_size(main_win_size[0], main_win_size[1])
        self.mainwindow.connect("destroy", self._on_mainwindow_destroy)
        self.mainwindow.connect(
            "size-allocate", self._on_mainwindow_size_allocate
        )

        try:
            self.mainwindow.set_icon_list([
                self.core.call_success("icon_get_pixbuf", "paperwork", 16),
                self.core.call_success("icon_get_pixbuf", "paperwork", 32),
                self.core.call_success("icon_get_pixbuf", "paperwork", 48),
                self.core.call_success("icon_get_pixbuf", "paperwork", 64),
                self.core.call_success("icon_get_pixbuf", "paperwork", 128),
            ])
        except KeyError as exc:
            # Will fail when generating data for the first time.
            LOGGER.warning("Failed to load main window icon", exc_info=exc)

        self.stacks = {
            "left": {
                "header": self.widget_tree.get_object(
                    "mainwindow_stack_header_left"
                ),
                "body": self.widget_tree.get_object(
                    "mainwindow_stack_body_left"
                ),
            },
            "right": {
                "header": self.widget_tree.get_object(
                    "mainwindow_stack_header_right"
                ),
                "body": self.widget_tree.get_object(
                    "mainwindow_stack_body_right"
                ),
            },
        }

        self.widget_tree.get_object("mainwindow_box_titlebar").connect(
            "notify::folded", self._on_titlebar_fold_change
        )
        self._on_titlebar_fold_change()

    def chkdeps(self, out: dict):
        if not GLIB_AVAILABLE:
            out['glib'].update(openpaperwork_core.deps.GLIB)
        if not GTK_AVAILABLE:
            out['gtk'].update(openpaperwork_gtk.deps.GTK)

    def on_gtk_initialized(self):
        app = self.core.call_success("gtk_get_app")
        # when remote, this segfaults...
        if not app.get_is_remote():
            self.mainwindow.set_application(app)

        for side in self.defaults.keys():
            self.mainwindow_show_default(side)

        self.mainwindow.set_visible(True)
        self.core.call_all("on_gtk_window_opened", self.mainwindow)

    def on_quit(self):
        # needed to save window size
        # TODO(JFlesch): not really config --> should not be stored in config ?
        if self.mainwindow is not None:
            self.core.call_all("config_save")
            self.mainwindow.set_visible(False)

    def _on_titlebar_fold_change(self, *args, **kwargs):
        LOGGER.info("Main window fold has changed")
        self.core.call_all("on_mainwindow_fold_change")

    def mainwindow_get_folded(self):
        return self.widget_tree.get_object(
            "mainwindow_box_titlebar"
        ).get_folded()

    def _on_mainwindow_destroy(self, main_window):
        LOGGER.info("Main window destroy. Quitting")
        self.core.call_all("on_gtk_window_closed", self.mainwindow)
        self.core.call_all("mainloop_quit_graceful")

    def _on_mainwindow_size_allocate(self, main_win, rectangle):
        (w, h) = main_win.get_size()
        if self._mainwindow_size == (w, h):
            return
        self._mainwindow_size = (w, h)
        self.core.call_all("config_put", "main_window_size", (w, h))

    def mainwindow_get_main_container(self):
        return self.widget_tree.get_object("main_box")

    def mainwindow_add(self, side: str, name: str, prio: int, header, body):
        self.components[side][name] = {
            "header": header,
            "body": body,
        }
        components = self.components[side][name]
        stacks = self.stacks[side]
        for (position, widget) in components.items():
            if widget is None:
                continue
            stacks[position].add_named(widget, name)
        self.defaults[side].append((prio, name))
        self.defaults[side].sort(reverse=True)
        return True

    def mainwindow_show(self, side: str, name: Optional[str] = None):
        self.widget_tree.get_object(
            "mainwindow_box_titlebar"
        ).set_visible_child(
            self.stacks[side]['header']
        )
        self.widget_tree.get_object(
            "mainwindow_box_body"
        ).set_visible_child(
            self.stacks[side]['body']
        )

        if name is None:
            return False

        if name in self.navigation_stacks[side]:
            self.navigation_stacks[side].remove(name)
        self.navigation_stacks[side].append(name)
        return self._mainwindow_show(side)

    def _mainwindow_show(self, side: str):
        stacks = self.stacks[side]

        LOGGER.info("Navigation: %s", self.navigation_stacks[side])

        # some plugin give us None as component, which means to use
        # the previous component in the navigation stack (see doclist
        # in multiple selection mode)
        component_names = {h: None for h in stacks.keys()}
        for name in reversed(self.navigation_stacks[side]):
            LOGGER.info("Showing %s on %s", name, side)
            for (k, v) in self.components[side][name].items():
                if component_names[k] is None and v is not None:
                    component_names[k] = name
            if None not in component_names.values():
                break

        has_changed = False
        for (h, stack) in stacks.items():
            if h not in component_names or component_names[h] is None:
                continue
            if stack.get_visible_child_name() == component_names[h]:
                continue
            has_changed = True
            stack.set_visible_child_name(component_names[h])
        return has_changed

    def mainwindow_show_default(self, side: str):
        # the default component may be incomplete (for instance, it may
        # provide a header, but not a body). So we update the navigation
        # stack to make it the behaviour consistent:
        # We ensure it contains at least one element with all components
        # We respect priorities

        for (prio, name) in self.defaults[side]:
            self.navigation_stacks[side].append(name)
            if None not in self.components[side][name].values():
                break
        first = self.navigation_stacks[side].pop(0)
        self.navigation_stacks[side].reverse()

        self.core.call_all("mainwindow_show", side, first)

    def mainwindow_back(self, side: str):
        navigation_stack = self.navigation_stacks[side][:-1]
        self.navigation_stacks[side] = navigation_stack
        self._mainwindow_show(side)

    def app_actions_add(self, action):
        if self.mainwindow is not None:
            self.mainwindow.add_action(action)

    def mainwindow_focus(self):
        self.mainwindow.grab_focus()

    def mainwindow_present(self):
        if self.mainwindow is not None:
            self.mainwindow.present()
        else:
            LOGGER.warn("Activated a second time but mainwindow is None")

    def _draw_overlay(self, img, area, color):
        overlay = PIL.Image.new('RGBA', img.size, color + (0,))
        draw = PIL.ImageDraw.Draw(overlay)
        draw.rectangle(area, fill=color + (0x5F,))
        return PIL.Image.alpha_composite(img, overlay)

    def screenshot_snap_all_doc_widgets(self, out_dir):
        out = self.core.call_success("fs_join", out_dir, "main_window.png")
        self.core.call_success(
            "screenshot_snap_widget", self.mainwindow, out
        )

        left = self.stacks['left']['body']
        left_alloc = left.get_allocation()
        left_width = left_alloc.width
        (left_x, left_y) = left.translate_coordinates(self.mainwindow, 0, 0)
        left = left_x + left_width

        with self.core.call_success("fs_open", out, 'rb') as fd:
            img = PIL.Image.open(fd)
            img.load()
            img = img.convert("RGBA")

        img = self._draw_overlay(
            img, (0, 0, left, img.size[1]), (0, 0xFF, 0)
        )
        img = self._draw_overlay(
            img, (left, 0, img.size[0], img.size[1]), (0, 0, 0xFF)
        )

        # drop some black border in the scn
        img = img.crop(
            (left_x, left_x, img.size[0] - left_x, img.size[1] - left_x)
        )

        # keep only the top 150px
        img = img.crop((0, 0, img.size[0], min(300, img.size[1])))

        out = self.core.call_success(
            "fs_join", out_dir, "main_window_split.png"
        )
        with self.core.call_success("fs_open", out, 'wb') as fd:
            img.save(fd, format="PNG")

    def app_shortcut_add(
            self, shortcut_group, shortcut_desc, shortcut_keys, action_name):
        """
        Arguments:
           - shortcut_group: Group name in which this shortcut belongs
             (human readable, translated ; actually used in the shortcut window
             if enabled)
           - shortcut_desc: Describe the shortcut
             (human readable, translated ; actually used in the shortcut window
             if enabled)
           - shortcut_keys: see gtk_accelerator_parse()
           - action_name: GAction that must be triggered
        """
        if self.mainwindow is None or not GTK_AVAILABLE:
            return
        LOGGER.info("Keyboard shortcut: %s --> %s", shortcut_keys, action_name)
        Gtk.Application.get_default().set_accels_for_action(
            action_name, (shortcut_keys, None)
        )



#    Paperwork - Using OCR to grep dead trees the easy way
#    Copyright (C) 2012-2019  Jerome Flesch
#
#    Paperwork is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Paperwork is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with Paperwork.  If not, see <http://www.gnu.org/licenses/>.
import logging

import openpaperwork_core
import openpaperwork_core.promise

import rich.tree

from .. import _


LOGGER = logging.getLogger(__name__)

DEFAULT_RESOLUTION = 300


class Plugin(openpaperwork_core.PluginBase):
    def __init__(self):
        super().__init__()
        self.out = []

    def get_interfaces(self):
        return ['shell']

    def get_deps(self):
        return [
            {
                "interface": "mainloop",
                "defaults": ["openpaperwork_gtk.mainloop.glib"],
            },
            {
                "interface": "config",
                "defaults": ['openpaperwork_core.config'],
            },
            {
                "interface": "scan",
                "defaults": ['paperwork_backend.docscan.libinsane'],
            },
        ]

    def cmd_complete_argparse(self, parser):
        scanner_parser = parser.add_parser(
            'scanner', help=_("Manage scanner configuration")
        )
        subparser = scanner_parser.add_subparsers(
            help=_("sub-command"), dest='subcommand', required=True
        )

        subparser.add_parser(
            'list', help=_("List all scanners and their possible settings")
        )

        subparser.add_parser(
            'get', help=_(
                "Show the currently selected scanner and its settings"
            )
        )

        set_scanner = subparser.add_parser(
            'set', help=_("Define which scanner and which settings to use")
        )
        set_scanner.add_argument(
            "device_id", help=_("Scanner to use")
        )
        set_scanner.add_argument(
            "--source", "-s", type=str, required=False,
            help=_(
                "Default source on the scanner to use (if not specified,"
                " one will be selected randomly)"
            )
        )
        set_scanner.add_argument(
            "--resolution", "-r", type=int, required=False,
            help=_("Default resolution (dpi ; default=300)")
        )

    def _get_scanner_info(self, dev, console, dev_id, dev_name):
        console.print(_("Examining scanner {} ...").format(dev_id))
        dev_out = {
            'id': dev_id,
            'name': dev_name,
            'sources': [],
        }
        sources = dev.get_sources()
        for (source_id, source) in sources.items():
            source_out = {
                "id": source_id,
                "resolutions": source.get_resolutions(),
            }
            dev_out['sources'].append(source_out)
        self.out.append(dev_out)
        return dev

    def _on_get_scanner_error(self, exception, scanner_dev_id):
        LOGGER.error(
            "Error while looking for options of %s",
            scanner_dev_id, exc_info=exception
        )

    def _get_scanners_info(self, devs, console):
        for dev in devs:
            # Schedule the promises to examine each scanner separately,
            # so if one failed (for instance on Windows when a scanner is
            # not actually connected), the others can still run.
            dev_id = dev[0]
            dev_name = dev[1]
            promise = openpaperwork_core.promise.Promise(
                self.core, hide_caught_exceptions=True
            )
            promise = promise.then(self.core.call_success(
                "scan_get_scanner_promise", dev_id
            ))
            promise.catch(self._on_get_scanner_error, dev_id)
            promise = promise.then(
                self._get_scanner_info, console, dev_id, dev_name
            )
            promise = promise.then(lambda scanner: scanner.close())
            promise = promise.then(lambda *args, **kwargs: None)
            self.core.call_success("scan_schedule", promise)

    def _list_scanners(self, console):
        self.out = []

        promise = self.core.call_success("scan_list_scanners_promise")
        promise = promise.then(self._get_scanners_info, console)
        self.core.call_success("scan_schedule", promise)
        self.core.call_all("mainloop_quit_graceful")
        self.core.call_one("mainloop")

    def _print_scanners(self, console):
        console.print("")
        tree = rich.tree.Tree("Scanners")
        for dev in self.out:
            dev_tree = tree.add(dev['name'])
            dev_tree.add(_("ID:") + " " + dev['id'])
            for source in dev['sources']:
                source_tree = dev_tree.add(_("Source:") + " " + source['id'])
                source_tree.add(
                    _("Resolutions:")
                    + " "
                    + str(source['resolutions'])
                )
        console.print(tree)

    def _get_scanner(self, console):
        out = {
            'id': self.core.call_success(
                "config_get", "scanner_dev_id"
            ),
            'source': self.core.call_success(
                "config_get", "scanner_source_id"
            ),
            'resolution': self.core.call_success(
                "config_get", "scanner_resolution"
            ),
        }
        console.print(_("ID:") + " " + str(out['id']))
        console.print(_("Source:") + " " + str(out['source']))
        console.print(_("Resolution:") + " " + str(out['resolution']))
        return out

    def _set_scanner(self, console, args):
        dev_settings = {
            'id': args.device_id,
            'source': args.source,
            'resolution': (
                args.resolution
                if args.resolution is not None
                else DEFAULT_RESOLUTION
            ),
        }

        # In any case, we want to make sure the settings provided are valid
        promise = self.core.call_success(
            "scan_get_scanner_promise", args.device_id
        )

        def check_source(dev):
            sources = dev.get_sources()

            if (
                        dev_settings['source'] is not None
                        and dev_settings['source'] not in sources
                    ):
                console.print(_(
                    "Source {} not found on device."
                    " Using another source"
                ).format(dev_settings['source']))
                dev_settings['source'] = None

            if dev_settings['source'] is None:
                for (source_id, source_obj) in sources.items():
                    if 'flatbed' in source_id.lower():
                        source = (source_id, source_obj)
                        break
                else:
                    source = sources.popitem()
                dev_settings['source'] = source[0]
            else:
                source = (
                    dev_settings['source'],
                    sources[dev_settings['source']]
                )

            console.print(_("Default source:") + " " + dev_settings['source'])

            return source[1]

        promise = promise.then(check_source)

        def check_resolution(source):
            resolutions = source.get_resolutions()
            if dev_settings['resolution'] in resolutions:
                return source
            resolution = min(
                resolutions, key=lambda x: abs(x - dev_settings['resolution'])
            )
            console.print(
                _("Resolution {} not available. Adjusted to {}.").format(
                    dev_settings['resolution'], resolution
                )
            )
            dev_settings['resolution'] = resolution
            return source

        promise = promise.then(check_resolution)

        promise = promise.then(lambda source: source.close())
        self.core.call_success("scan_schedule", promise)
        self.core.call_all("mainloop_quit_graceful")
        self.core.call_one("mainloop")

        self.core.call_all(
            "config_put", "scanner_dev_id", dev_settings['id']
        )
        self.core.call_all(
            "config_put", "scanner_source_id", dev_settings['source']
        )
        self.core.call_all(
            "config_put", "scanner_resolution",
            dev_settings['resolution']
        )
        self.core.call_all("config_save")

        return self._get_scanner(console)

    def cmd_run(self, console, args):
        if args.command != 'scanner':
            return None

        if args.subcommand == 'list':
            console.print("Looking for scanners ...")
            self._list_scanners(console)
            console.print(f"{len(self.out)} scanners found")
            console.print("")
            self._print_scanners(console)
            return self.out
        elif args.subcommand == 'set':
            return self._set_scanner(console, args)
        elif args.subcommand == 'get':
            return self._get_scanner(console)
        assert False



import gettext


def _(s):
    return gettext.dgettext('openpaperwork_gtk', s)


CLI_PLUGINS = [
    # plugins that can make sense to use even in the CLI application
    'openpaperwork_gtk.external_apps.gio',
    'openpaperwork_gtk.fs.gio',
    'openpaperwork_gtk.l10n',
    'openpaperwork_gtk.mainloop.glib',
]

GUI_PLUGINS = [
    'openpaperwork_gtk.bug_report',
    'openpaperwork_gtk.bug_report.http',
    'openpaperwork_gtk.bug_report.zip',
    'openpaperwork_gtk.busy.mouse',
    'openpaperwork_gtk.colors',
    'openpaperwork_gtk.dialogs.single_entry',
    'openpaperwork_gtk.dialogs.yes_no',
    'openpaperwork_gtk.pixbuf.pillow',
    'openpaperwork_gtk.resources',
    'openpaperwork_gtk.screenshots',
    'openpaperwork_gtk.uncaught_exception',
    'openpaperwork_gtk.wait_for_background_tasks',
    'openpaperwork_gtk.widgets.calendar',
    'openpaperwork_gtk.widgets.progress',
]



import collections
import logging

try:
    from gi.repository import GObject
    GLIB_AVAILABLE = True
except (ImportError, ValueError):
    GLIB_AVAILABLE = False

try:
    import gi
    gi.require_version('Gdk', '3.0')
    gi.require_version('Gtk', '3.0')
    from gi.repository import Gdk
    from gi.repository import Gtk
    GTK_AVAILABLE = True
except (ImportError, ValueError):
    # workaround so chkdeps can still be called
    class Gtk(object):
        class Box(object):
            pass

        class Widget(object):
            pass

    GTK_AVAILABLE = False


import openpaperwork_core
import openpaperwork_core.deps
import openpaperwork_gtk.deps


LOGGER = logging.getLogger(__name__)


class WidgetInfo(object):
    def __init__(self, widget, alignment, size=None):
        self.widget = widget
        self.alignment = alignment
        if size is not None:
            self.size = (int(size[0]), int(size[1]))
        else:
            self.size = (0, 0)
        self.position = (-1, -1)
        self.visible = False

    def update_widget_size(self):
        if self.widget is None:  # test mode
            return
        self.size = (
            self.widget.get_preferred_width()[0],
            self.widget.get_preferred_height()[0],
        )

    def is_visible(self):
        if self.widget is None:  # test mode
            return True
        return self.widget.get_visible()

    def __eq__(self, o):
        if self is o:
            return True
        return self.widget == o


def recompute_height_for_width(widgets, width, spacing=(0, 0)):
    height = spacing[1]
    line_height = 0
    line_width = 0
    for widget in widgets:
        if not widget.is_visible():
            continue
        if widget.size == (0, 0):
            widget.update_widget_size()
        if line_width + widget.size[0] + spacing[0] > width + 1:
            height += spacing[1]
            height += line_height
            line_width = 0
            line_height = 0
        line_width += spacing[0]
        line_width += widget.size[0]
        line_height = max(line_height, widget.size[1])
    height += line_height
    height += spacing[1]
    return height


def recompute_box_positions(core, widgets, width, spacing=(0, 0)):
    core.call_all("on_perfcheck_start", "recompute_box_positions")

    # build lines
    lines = []
    line_heights = []
    line = []
    line_width = spacing[0]
    max_line_width = 0
    line_height = 0
    for widget in widgets:
        if not widget.is_visible():
            continue
        widget.update_widget_size()
        if line_width + widget.size[0] + spacing[0] > width + 1:
            lines.append(line)
            line_heights.append(line_height)
            line = []
            max_line_width = max(max_line_width, line_width)
            line_width = 0
            line_height = 0
        line.append(widget)
        line_width += spacing[0]
        line_width += widget.size[0]
        line_height = max(line_height, widget.size[1])
    max_line_width = max(max_line_width, line_width)
    lines.append(line)
    line_heights.append(line_height)

    # sort widgets by alignment in each line
    def sort_widgets_per_alignments(widgets):
        out = {
            Gtk.Align.START: [],
            Gtk.Align.CENTER: [],
            Gtk.Align.END: [],
        }
        for widget in widgets:
            out[widget.alignment].append(widget)
        return out

    lines = [sort_widgets_per_alignments(line) for line in lines]

    # position widgets in lines
    height = spacing[1]
    for (line, line_height) in zip(lines, line_heights):
        nb_columns = 0
        w_start = 0
        w_end = width

        # start
        for widget in line[Gtk.Align.START]:
            nb_columns += 1
            w_start += spacing[0]
            widget.position = (w_start, height)
            w_start += widget.size[0]

        # end
        line[Gtk.Align.END].reverse()
        for widget in line[Gtk.Align.END]:
            nb_columns += 1
            w_end -= spacing[0]
            w_end -= widget.size[0]
            widget.position = (w_end, height)

        # center
        w_center = sum(w.size[0] for w in line[Gtk.Align.CENTER])
        w_center += spacing[0] * (len(line[Gtk.Align.CENTER]) - 1)
        w_center = (width - w_center) / 2
        w_orig = w_center
        for widget in line[Gtk.Align.CENTER]:
            nb_columns += 1
            if w_center != w_orig:
                w_center += spacing[0]
            widget.position = (int(w_center), int(height))
            w_center += widget.size[0]

        height += line_height
        height += spacing[1]

    core.call_all(
        "on_perfcheck_stop", "recompute_box_positions",
        nb_boxes=len(widgets)
    )
    return (widgets, max_line_width, height)


class CustomFlowLayout(Gtk.Box):
    def __init__(self, core, spacing=(0, 0)):
        super().__init__()
        self.core = core

        self.widgets = collections.OrderedDict()

        self.spacing = spacing
        self.vadjustment = None
        self.bottom_margin = 0
        self.allocation = None

        self.set_has_window(False)
        self.set_redraw_on_allocate(False)

        self.connect("size-allocate", self._on_size_allocate)
        self.connect("add", self._on_add)
        self.connect("remove", self._on_remove)

    def _on_add(self, _, widget):
        self.recompute_layout()
        self.queue_resize()

    def _on_remove(self, _, widget):
        self.widgets.pop(widget)
        self.queue_draw()
        self.queue_resize()

    def do_forall(self, include_internals: bool, callback, callback_data=None):
        if not hasattr(self, 'widgets'):
            return
        widgets = self.widgets.copy()
        for widget in widgets:
            callback(widget)

    def add_child(self, widget, alignment):
        w = WidgetInfo(widget, alignment)
        self.widgets[widget] = w
        self.add(widget)

    def do_get_request_mode(self):
        return Gtk.SizeRequestMode.WIDTH_FOR_HEIGHT

    def do_get_preferred_width(self):
        nat_width = 0
        for widget in self.widgets.values():
            widget.update_widget_size()
            nat_width += widget.size[0]
        return (0, nat_width)

    def do_get_preferred_height_for_width(self, width):
        requested_height = recompute_height_for_width(
            self.widgets.values(), width, self.spacing
        )
        requested_height += self.bottom_margin
        return (requested_height, requested_height)

    def do_get_preferred_height(self):
        (min_width, nat_width) = self.do_get_preferred_width()
        return self.do_get_preferred_height_for_width(min_width)

    def do_get_preferred_width_for_height(self, height):
        return self.do_get_preferred_width()

    def _on_size_allocate(self, _, allocation):
        self.allocation = allocation
        self.recompute_layout()

    def recompute_layout(self):
        if self.allocation is None:
            return

        for widget in self.widgets.values():
            widget.update_widget_size()

        (
            _, self.requested_width, self.requested_height
        ) = recompute_box_positions(
            self.core, self.widgets.values(), self.allocation.width,
            self.spacing
        )
        self.requested_height += self.bottom_margin

        for widget in self.widgets.values():
            rect = Gdk.Rectangle()
            rect.x = self.allocation.x + widget.position[0]
            rect.y = self.allocation.y + widget.position[1]
            rect.width = widget.size[0]
            rect.height = widget.size[1]
            widget.widget.size_allocate(rect)

    def _on_destroy(self, _):
        if not hasattr(self, 'widgets'):
            return
        for widget in self.widgets.keys():
            widget.unparent()
        self.widgets = collections.OrderedDict()

    def set_bottom_margin(self, height):
        self.bottom_margin = height
        self.queue_resize()


if GTK_AVAILABLE:
    GObject.type_register(CustomFlowLayout)


class Plugin(openpaperwork_core.PluginBase):
    def get_interfaces(self):
        return [
            'chkdeps',
            'gtk_widget_flowlayout',
        ]

    def chkdeps(self, out: dict):
        if not GLIB_AVAILABLE:
            out['glib'].update(openpaperwork_core.deps.GLIB)
        if not GTK_AVAILABLE:
            out['gtk'].update(openpaperwork_gtk.deps.GTK)

    def gtk_widget_flowlayout_new(self, spacing=(0, 0)):
        assert GLIB_AVAILABLE
        assert GTK_AVAILABLE
        return CustomFlowLayout(self.core, spacing)



import logging

from .. import PluginBase


LOGGER = logging.getLogger(__name__)


class Plugin(PluginBase):
    """
    Censor bug report attachments
    """
    PRIORITY = -10000

    def get_interfaces(self):
        return ['bug_report_attachments']

    def get_deps(self):
        return [
            {
                'interface': 'censor',
                'defaults': ['openpaperwork_core.censor'],
            },
        ]

    def _censor_attachment(self, attachment_id, args):
        url = self.core.call_success(
            "bug_report_get_attachment_file_url", attachment_id, *args
        )
        if url is None:
            LOGGER.info(
                "Attachment %s has no URL yet. Can't censor", attachment_id
            )
            return

        if not url.endswith(".conf") and not url.endswith(".txt"):
            LOGGER.info(
                    "Unknown file type: %s:%s. Can't censor",
                    attachment_id, url
            )
            return

        basename = self.core.call_success("fs_basename", url)
        if basename.startswith("censored_"):
            LOGGER.info("Attachmnent %s appears to be already censored", url)
            return

        LOGGER.info("Censoring %s:%s", attachment_id, url)

        censored = self.core.call_success(
            "censor_txt_file", url, tmp_on_disk=True
        )
        self.core.call_all(
            "bug_report_update_attachment", attachment_id, {
                "censored": True,
                "file_url": censored,
            }, *args
        )

    def on_bug_report_attachment_selected(self, attachment_id, *args):
        self._censor_attachment(attachment_id, args)

    def bug_report_update_attachment(self, attachment_id, infos: dict, *args):
        if 'censored' in infos:
            return
        self._censor_attachment(attachment_id, args)



#    Paperwork - Using OCR to grep dead trees the easy way
#    Copyright (C) 2012-2019  Jerome Flesch
#
#    Paperwork is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Paperwork is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with Paperwork.  If not, see <http://www.gnu.org/licenses/>.
import datetime
import logging

from .. import (_, PluginBase)


LOGGER = logging.getLogger(__name__)


def bool_from_str(s):
    s = s.strip()
    if s == "" or s == "0" or s.lower() == "false":
        return False
    return True


# Only basic types are handled by shell commands
CMD_VALUE_TYPES = {
    'str': str,
    'int': int,
    'float': float,
    'bool': bool_from_str,
}

DATE_FORMAT = "%Y-%m-%d"

TO_JSON = {
    datetime.date: lambda x: x.strftime(DATE_FORMAT)
}


class Plugin(PluginBase):
    def __init__(self):
        super().__init__()

    def get_interfaces(self):
        return ['shell']

    def get_deps(self):
        return [
            {
                'interface': 'config',
                'defaults': ['openpaperwork_core.config'],
            },
        ]

    def cmd_complete_argparse(self, parser):
        config_parser = parser.add_parser(
            'config', help=_("Manage Paperwork configuration")
        )

        subparser = config_parser.add_subparsers(
            help=_("sub-command"), dest='subcommand', required=True
        )

        get_parser = subparser.add_parser(
            'get', help=_("Get a value from Paperwork's configuration")
        )
        get_parser.add_argument('opt_name')

        put_parser = subparser.add_parser(
            'put', help=_("Set a value in Paperwork's configuration")
        )
        put_parser.add_argument('opt_name')
        put_parser.add_argument('type', help=_("See 'config list_type'"))
        put_parser.add_argument('value')

        subparser.add_parser(
            'show', help=_("Show Paperwork's configuration")
        )

        subparser.add_parser(
            'list_types', help=_(
                "Show value types you can use from command line"
            )
        )

    def cmd_run(self, console, args):
        if args.command != 'config':
            return None
        if args.subcommand == "get":
            return self._cmd_get(args.opt_name)
        elif args.subcommand == "put":
            return self._cmd_put(args.opt_name, args.type, args.value)
        elif args.subcommand == "show":
            return self._cmd_show()
        elif args.subcommand == "list_types":
            return self._cmd_list_types()
        else:
            return None

    def _cmd_get(self, opt_name):
        v = self.core.call_success("config_get", opt_name)
        if v is None:
            LOGGER.warning("No such option '%s'", opt_name)
            return None
        self.core.call_all("print", f"{opt_name} = {v}")
        self.core.call_success("print_flush")
        return {opt_name: v}

    def _cmd_put(self, opt_name, vtype, value):
        value = CMD_VALUE_TYPES[vtype](value)
        self.core.call_success("print", f"{opt_name} = {value}")
        self.core.call_success("print_flush")
        self.core.call_all("config_put", opt_name, value)
        self.core.call_all("config_save")
        return {opt_name: value}

    def _cmd_show(self):
        opts = self.core.call_success("config_list_options")
        out = {}
        opts.sort()
        for opt in opts:
            v = self.core.call_success("config_get", opt)
            if type(v) in TO_JSON:
                v = TO_JSON[type(v)](v)
            out[opt] = v
            self.core.call_success(
                "print", f"{opt} = {out[opt]} ({type(v).__name__})"
            )
        self.core.call_success("print_flush")
        return out

    def _cmd_list_types(self):
        r = list(CMD_VALUE_TYPES.keys())
        self.core.call_success("print", str(r))
        self.core.call_success("print_flush")
        return r



"""
Manages a configuration file using configparser.
"""

import collections
import configparser
import datetime
import logging

from ... import (_, PluginBase)


LOGGER = logging.getLogger(__name__)


class ConfigBool(object):
    def __init__(self, value=False):
        if isinstance(value, str):
            self.value = (value.lower() == 'true')
        else:
            self.value = value

    def __eq__(self, o):
        return (self.value == o)

    def __bool__(self):
        return self.value

    def __str__(self):
        return str(self.value)

    def get_value(self):
        return self.value


class ConfigDate(object):
    DATE_FORMAT = "%Y-%m-%d"

    def __init__(self, value=datetime.datetime(year=1971, month=1, day=1)):
        if isinstance(value, str):
            self.value = (
                datetime.datetime
                .strptime(value, self.DATE_FORMAT)
                .date()
            )
        else:
            self.value = value

    def __eq__(self, o):
        return (self.value == o)

    def __str__(self):
        return self.value.strftime(self.DATE_FORMAT)

    def get_value(self):
        return self.value


class ConfigList(object):
    SEPARATOR = ";"

    def __init__(self, value=None, elements=None):
        if elements is None:
            elements = []
        self.elements = elements

        if value is not None:
            if isinstance(value, str):
                if value != '':
                    elements = value.split(self.SEPARATOR)
                    for i in elements:
                        (t, v) = i.split(_TYPE_SEPARATOR, 1)
                        self.elements.append(_STR_TO_TYPE[t](v))
            elif hasattr(value, 'elements'):
                self.elements = value.elements[:]
            else:
                self.elements = list(value)

    def __iter__(self):
        return iter(self.elements)

    def __contains__(self, o):
        return o in self.elements

    def __getitem__(self, *args, **kwargs):
        return self.elements.__getitem__(*args, **kwargs)

    def __setitem__(self, *args, **kwargs):
        return self.elements.__setitem__(*args, **kwargs)

    def __len__(self):
        return len(self.elements)

    def append(self, value):
        self.elements.append(value)

    def remove(self, value):
        self.elements.remove(value)

    def __str__(self):
        out = []
        for e in self.elements:
            out.append("{}{}{}".format(
                _TYPE_TO_STR[type(e)], _TYPE_SEPARATOR, str(e)
            ))
        return self.SEPARATOR.join(out)

    def get_value(self):
        return [
            e.get_value()
            if hasattr(e, 'get_value')
            else e
            for e in self.elements
        ]


class ConfigDict(object):
    SEPARATOR_ITEMS = ";"
    SEPARATOR_KEYVALS = "="

    def __init__(self, value=None, elements={}):
        self.elements = elements

        if value is not None:
            if isinstance(value, str):
                elements = value.split(self.SEPARATOR_ITEMS)
                for i in elements:
                    (k, v) = i.split(self.SEPARATOR_KEYVALS, 1)
                    (t, v) = v.split(_TYPE_SEPARATOR, 1)
                    self.elements[k] = _STR_TO_TYPE[t](v)
            elif hasattr(value, 'elements'):
                self.elements = value.elements[:]
            else:
                self.elements = dict(value)

    def __iter__(self):
        return iter(self.elements)

    def __contains__(self, o):
        return o in self.elements

    def __getitem__(self, *args, **kwargs):
        return self.elements.__getitem__(*args, **kwargs)

    def __setitem__(self, *args, **kwargs):
        return self.elements.__setitem__(*args, **kwargs)

    def __len__(self):
        return len(self.elements)

    def __str__(self):
        out = []
        for (k, v) in self.elements.items():
            out.append("{}{}{}{}{}".format(
                k, self.SEPARATOR_KEYVALS,
                _TYPE_TO_STR[type(v)], _TYPE_SEPARATOR, str(v)
            ))
        return self.SEPARATOR_ITEMS.join(out)

    def get_value(self):
        return {
            k: v.get_value()
            if hasattr(v, 'get_value')
            else v
            for (k, v) in self.elements.items()
        }


_TYPE_TO_STR = {
    bool: "bool",
    ConfigBool: "bool",
    ConfigDate: "date",
    ConfigDict: "dict",
    ConfigList: "list",
    datetime.date: "date",
    dict: "dict",
    float: "float",
    int: "int",
    list: "list",
    str: "str",
    tuple: "list",
}
_STR_TO_TYPE = {
    "bool": ConfigBool,
    "date": ConfigDate,
    "dict": ConfigDict,
    "list": ConfigList,
    "float": float,
    "int": int,
    "str": str,
}
_TYPE_SEPARATOR = ":"


class Plugin(PluginBase):
    TEST_FILE_URL = None  # unit tests only

    def __init__(self):
        self.config = configparser.RawConfigParser()
        self.base_path = None
        self.application_name = None
        self.config_path = None
        self.observers = collections.defaultdict(set)
        self.core = None
        self.default_plugins = []

    def get_interfaces(self):
        return ['config_backend']

    def get_deps(self):
        return [
            {
                'interface': 'fs',
                'defaults': ['openpaperwork_core.fs.python'],
            },
            {
                'interface': 'paths',
                'defaults': ['openpaperwork_core.paths.xdg'],
            },
        ]

    def init(self, core):
        self.core = core
        if self.base_path is None:
            self.base_path = self.core.call_success("paths_get_config_dir")

    def _get_filepath(self):
        if self.TEST_FILE_URL is not None:
            return self.TEST_FILE_URL
        return self.core.call_success(
            "fs_join", self.base_path, self.application_name + ".conf"
        )

    def config_backend_load(self, application_name):
        self.application_name = application_name
        self.config_path = self._get_filepath()
        self.config = configparser.RawConfigParser()
        LOGGER.info("Loading configuration '%s' ...", self.config_path)
        if self.core.call_success("fs_exists", self.config_path) is not None:
            fd = self.core.call_success("fs_open", self.config_path, 'r')
            with fd:
                self.config.read_file(fd)
        else:
            LOGGER.warning(
                "Cannot load configuration '%s'. File does not exist",
                self.config_path
            )
        for observers in self.observers.values():
            for observer in observers:
                observer()

    def config_backend_save(self, application_name=None):
        if application_name is not None:
            self.application_name = application_name
        config_path = self._get_filepath()
        LOGGER.info("Writing configuration '%s' ...", config_path)
        with self.core.call_success("fs_open", config_path, 'w') as fd:
            self.config.write(fd)

    def config_backend_load_plugins(self, opt_name, default=[]):
        """
        Load and init the plugin list from the configuration.
        """
        self.default_plugins = default
        modules = self.config_backend_get(
            "plugins", opt_name, ConfigList(None, self.default_plugins)
        )
        LOGGER.info(
            "Loading and initializing plugins from configuration: %s",
            str(modules)
        )
        for module in modules:
            self.core.load(module)
        self.core.init()

    def config_backend_list_active_plugins(self, opt_name):
        return self.config_backend_get(
            "plugins", opt_name, ConfigList(None, self.default_plugins)
        )

    def config_backend_reset_plugins(self, opt_name):
        self.config_backend_del("plugins", opt_name)

    def config_backend_add_plugin(self, opt_name, module_name):
        LOGGER.info("Adding plugin '%s' to configuration", module_name)
        modules = self.config_backend_list_active_plugins(opt_name)
        modules.append(module_name)
        self.config_backend_put("plugins", opt_name, modules)

    def config_backend_remove_plugin(self, opt_name, module_name):
        LOGGER.info("Removing plugin '%s' from configuration", module_name)
        modules = self.config_backend_list_active_plugins(opt_name)
        try:
            modules.remove(module_name)
        except ValueError:
            LOGGER.warning("Plugin '%s' not found", module_name)
        self.config_backend_put("plugins", opt_name, modules)

    def config_backend_put(self, section, key, value):
        """
        Section must be a string.
        Key must be a string.
        """
        LOGGER.debug("Configuration: %s:%s <-- %s", section, key, str(value))
        if value is None:
            if section not in self.config:
                return
            if key not in self.config[section]:
                return
            self.config[section].pop(key)
            return
        t_str = _TYPE_TO_STR[type(value)]
        t = _STR_TO_TYPE[t_str]
        value = t(value)
        value = "{}{}{}".format(t_str, _TYPE_SEPARATOR, str(value))
        if section not in self.config:
            self.config[section] = {key: value}
        else:
            self.config[section][key] = value

        for observer in self.observers[section]:
            observer()

    def config_backend_del(self, section, key):
        if section not in self.config:
            return
        if key not in self.config[section]:
            return
        self.config.remove_option(section, key)

    def config_backend_get(self, section, key, default=None):
        try:
            value = self.config[section][key]
            if value.strip() == "":
                return None
            (t, value) = value.split(_TYPE_SEPARATOR, 1)
            r = _STR_TO_TYPE[t](value)
            if hasattr(r, 'get_value'):
                r = r.get_value()
            LOGGER.debug("Configuration: %s:%s --> %s", section, key, str(r))
            return r
        except KeyError:
            LOGGER.debug(
                "Configuration: %s:%s --> %s (default value)",
                section, key, str(default)
            )
            return default

    def config_backend_add_observer(self, section, callback):
        self.observers[section].add(callback)

    def config_backend_remove_observer(self, section, callback):
        self.observers[section].remove(callback)

    def bug_report_get_attachments(self, out: dict):
        if self.config_path is None:
            return
        if self.core.call_success("fs_exists", self.config_path) is None:
            return
        out['config'] = {
            'include_by_default': True,
            'date': None,
            'file_type': _("App. config."),
            'file_url': self.config_path,
            'file_size': self.core.call_success("fs_getsize", self.config_path)
        }



import json
import logging

import openpaperwork_core


from .. import _


LOGGER = logging.getLogger(__name__)


class Plugin(openpaperwork_core.PluginBase):
    def get_interfaces(self):
        return ['authors']

    def authors_get(self, out: dict):
        try:
            translators = json.loads(
                # Translators: put your names here. See French translation
                # for reference.
                # Must valid JSON. Must be a list of strings (your names)
                _("[]")
            )
        except json.JSONDecodeError as exc:
            LOGGER.error("Failed to load translator list", exc_info=exc)
            return
        translators = [('', translator, -1) for translator in translators]
        out['Translators'] = translators



import logging

try:
    from gi.repository import Gio
    GIO_AVAILABLE = True
except (ImportError, ValueError):
    GIO_AVAILABLE = False

import openpaperwork_core
import openpaperwork_core.deps


LOGGER = logging.getLogger(__name__)


class Plugin(openpaperwork_core.PluginBase):
    def get_interfaces(self):
        return [
            'chkdeps',
            'action',
            'action_app',
            'action_app_find',
        ]

    def get_deps(self):
        return [
            {
                'interface': 'app_actions',
                'defaults': ['paperwork_gtk.mainwindow.window'],
            },
            {
                'interface': 'gtk_search_field',
                'defaults': ['paperwork_gtk.mainwindow.search.field'],
            },
        ]

    def chkdeps(self, out: dict):
        if not GIO_AVAILABLE:
            out['glib'].update(openpaperwork_core.deps.GLIB)

    def init(self, core):
        super().init(core)
        if not GIO_AVAILABLE:
            return
        action = Gio.SimpleAction.new('app_find', None)
        action.connect("activate", self._focus_on_search_field)
        self.core.call_all("app_actions_add", action)

    def _focus_on_search_field(self, action, parameter):
        self.core.call_all("search_focus")



import logging

try:
    import gi
    gi.require_version('Gtk', '3.0')
    from gi.repository import Gtk
    GTK_AVAILABLE = True
except (ImportError, ValueError):
    GTK_AVAILABLE = False

import openpaperwork_core
import openpaperwork_core.promise

from .. import (_, deps)


LOGGER = logging.getLogger(__name__)

PATH_MAKE_BUG_REPORT = "/beacon/bug_report/create"
PATH_ADD_ATTACHMENT = "/beacon/bug_report/add_attachment"


class Sender(object):
    def __init__(self, core, http, progress, description, file_urls):
        self.core = core
        self.http = http
        self.progress = progress

        self.description = description
        self.file_urls = file_urls
        self.infos = None

        self.nb_steps = len(self.file_urls) + 2
        self.current_step = 0

    def _notify_progress(self, *args, **kwargs):
        assert self.current_step <= self.nb_steps

        if self.current_step == 0:
            txt = _("Creating bug report ...")
        else:
            txt = _("Sending bug report attachment ...")
        self.core.call_all(
            "on_progress", "bug_report_http",
            self.current_step / self.nb_steps, txt
        )
        self.progress.set_fraction(self.current_step / self.nb_steps)
        self.progress.set_text(txt)
        self.current_step += 1

    def _set_report_infos(self, infos):
        self.infos = infos

    def _str_to_http(self, file_name, string):
        data = {
            "file_name": file_name,
            "binary": string.encode("utf-8")
        }
        data.update(self.infos)
        return data

    def _file_to_http(self, file_url):
        with self.core.call_success("fs_open", file_url, "rb") as fd:
            data = {
                "file_name": self.core.call_success("fs_basename", file_url),
                "binary": fd.read()
            }
        data.update(self.infos)
        return data

    def get_promise(self):
        LOGGER.info("Will send %s", self.file_urls)

        promise = openpaperwork_core.promise.Promise(
            self.core, self._notify_progress
        )

        promise = promise.then(lambda *args, **kwargs: {
            "app_name": self.core.call_success("app_get_name"),
            "app_version": self.core.call_success("app_get_version"),
        })
        promise = promise.then(self.http.get_request_promise(
            PATH_MAKE_BUG_REPORT
        ))
        promise = promise.then(self._set_report_infos)
        promise = promise.then(self._notify_progress)

        promise = promise.then(
            self._str_to_http, "description.txt", self.description
        )
        promise = promise.then(self.http.get_request_promise(
            PATH_ADD_ATTACHMENT
        ))
        promise = promise.then(self._notify_progress)

        for file_url in self.file_urls:
            promise = promise.then(self._file_to_http, file_url)
            promise = promise.then(self.http.get_request_promise(
                PATH_ADD_ATTACHMENT
            ))
            promise = promise.then(self._notify_progress)

        return promise

    def get_report_url(self):
        return self.infos['url']


class Plugin(openpaperwork_core.PluginBase):
    PRIORITY = 100

    def __init__(self):
        super().__init__()
        self.widget_tree = None
        self.assistant = None
        self.file_urls = []
        self.http = None
        self.progress = None
        self.prepare_handler_id = None
        self.description = None

    def get_interfaces(self):
        return [
            'bug_report_method',
            'chkdeps',
        ]

    def get_deps(self):
        return [
            {
                'interface': 'app',
                'defaults': [],
            },
            {
                'interface': 'gtk_resources',
                'defaults': ['openpaperwork_gtk.resources'],
            },
            {
                'interface': 'http_json',
                'defaults': ['openpaperwork_core.http'],
            },
            {
                'interface': 'mainloop',
                'defaults': ['openpaperwork_core.mainloop.asyncio'],
            },
        ]

    def init(self, core):
        super().init(core)
        self.http = self.core.call_success(
            "http_json_get_client", "bug_report"
        )

    def chkdeps(self, out: dict):
        if not GTK_AVAILABLE:
            out['gtk'].update(deps.GTK)

    def bug_report_complete(self, *args):
        self.core.call_all(
            "bug_report_add_method",
            _("Send automatically"),
            _("Send the bug report automatically to OpenPaper.work"),
            self._enable_http, self._disable_http,
            *args
        )

    def _enable_http(self, bug_report_widget_tree):
        self.widget_tree = self.core.call_success(
            "gtk_load_widget_tree",
            "openpaperwork_gtk.bug_report", "bug_report_http.glade"
        )

        assistant = bug_report_widget_tree.get_object("bug_report_dialog")
        self.assistant = assistant

        page = self.widget_tree.get_object("bug_report_http_description_page")
        description_page = page
        assistant.append_page(page)
        assistant.set_page_complete(page, False)
        self.description = self.widget_tree.get_object(
            "bug_report_http_description_buffer"
        )
        self.description.connect(
            "changed", lambda txt_buffer: assistant.set_page_complete(
                description_page, txt_buffer.get_char_count() > 0
            )
        )

        page = self.widget_tree.get_object("bug_report_http_progress")
        self.progress = page
        assistant.append_page(page)
        assistant.set_page_complete(page, False)

        page = self.widget_tree.get_object("bug_report_http_result")
        self.page_idx = assistant.append_page(page)
        assistant.set_page_complete(page, True)
        assistant.set_page_type(page, Gtk.AssistantPageType.CONFIRM)

        self.prepare_handler_id = assistant.connect(
            "prepare", self._on_prepare
        )

    def _disable_http(self, bug_report_widget_tree):
        if self.widget_tree is None:
            return
        assistant = bug_report_widget_tree.get_object("bug_report_dialog")
        assistant.remove(self.widget_tree.get_object(
            "bug_report_http_description_page"
        ))
        assistant.remove(self.widget_tree.get_object(
            "bug_report_http_progress"
        ))
        assistant.remove(self.widget_tree.get_object(
            "bug_report_http_result"
        ))
        assistant.disconnect(self.prepare_handler_id)
        self.widget_tree = None

    def bug_report_set_file_urls_to_send(self, file_urls):
        self.file_urls = file_urls

    def _on_prepare(self, assistant, page):
        if page != self.progress:
            return
        self._send()

    def _get_description(self):
        author = self.widget_tree.get_object("bug_report_http_author")
        author = author.get_text()

        start = self.description.get_iter_at_offset(0)
        end = self.description.get_iter_at_offset(-1)
        description = self.description.get_text(start, end, False)

        return "Author: {}\n\n{}".format(author, description)

    def _send(self):
        promise = openpaperwork_core.promise.Promise(
            self.core, self.core.call_all, args=("on_busy",)
        )
        sender = Sender(
            self.core, self.http, self.progress,
            self._get_description(), self.file_urls
        )
        promise = promise.then(sender.get_promise())
        promise = promise.then(self._show_result, sender)
        promise = promise.catch(self._on_error)
        promise.schedule()

    def _show_result(self, sender):
        self.core.call_all("on_idle")

        if self.widget_tree is None:
            return
        LOGGER.info("Transfer successful")
        url = self.widget_tree.get_object("bug_report_http_url")
        url.set_markup('<a href="{url}">{url}</a>'.format(
            url=sender.get_report_url(),
        ))
        self.assistant.set_page_complete(self.progress, True)
        self.progress.set_fraction(1.0)
        self.progress.set_text(_("Success"))

        self.assistant.set_current_page(
            # no point in staying on the progress bar page
            self.assistant.get_current_page() + 1
        )

    def _on_error(self, exc):
        self.core.call_all("on_idle")

        if self.widget_tree is None:
            return

        LOGGER.error("Transfer failed", exc_info=exc)
        exc_txt = str(exc)[:256]
        txt = self.widget_tree.get_object("bug_report_http_result_txt")
        txt.set_text(_("Transfer failed:\n\n%s") % exc_txt)
        url = self.widget_tree.get_object("bug_report_http_url")
        url.set_text("")
        self.assistant.set_page_complete(self.progress, True)

        self.progress.set_fraction(1.0)
        self.progress.set_text(_("FAILED"))
        self.core.call_all("on_progress", "bug_report_http", 1.0)



"""
Draw a Pillow image on top of GtkDrawingArea.
"""

import logging

import openpaperwork_core


LOGGER = logging.getLogger(__name__)


class Drawer(object):
    BACKGROUND = (0.75, 0.75, 0.75)

    def __init__(self, core, drawing_area, pil_img):
        self.core = core
        self.drawing_area = drawing_area  # expected: Gtk.DrawingArea
        self.img = core.call_success("pillow_to_surface", pil_img)
        self.draw_connect_id = drawing_area.connect("draw", self.on_draw)

    def stop(self):
        if self.draw_connect_id is not None:
            self.drawing_area.disconnect(self.draw_connect_id)
        self.draw_connect_id = None
        self.img = None  # free the memory

    def on_draw(self, drawing_area, cairo_ctx):
        widget_height = self.drawing_area.get_allocated_height()
        widget_width = self.drawing_area.get_allocated_width()
        factor_w = self.img.surface.get_width() / widget_width
        factor_h = self.img.surface.get_height() / widget_height
        factor = max(factor_w, factor_h)

        # background
        cairo_ctx.save()
        try:
            cairo_ctx.set_source_rgb(
                self.BACKGROUND[0], self.BACKGROUND[1], self.BACKGROUND[2]
            )
            cairo_ctx.rectangle(0, 0, widget_width, widget_height)
            cairo_ctx.clip()
            cairo_ctx.paint()
        finally:
            cairo_ctx.restore()

        # image
        cairo_ctx.save()
        try:
            cairo_ctx.scale(1.0 / factor, 1.0 / factor)
            cairo_ctx.set_source_surface(self.img.surface)
            cairo_ctx.rectangle(
                0, 0,
                self.img.surface.get_width(), self.img.surface.get_height()
            )
            cairo_ctx.clip()
            cairo_ctx.paint()
        finally:
            cairo_ctx.restore()


class Plugin(openpaperwork_core.PluginBase):
    def __init__(self):
        super().__init__()
        # drawing area --> Drawer
        self.active_drawers = {}

    def get_interfaces(self):
        return [
            'gtk_drawer_pillow',
        ]

    def get_deps(self):
        return [
            {
                'interface': 'pillow_to_surface',
                'defaults': ['paperwork_backend.cairo.pillow'],
            },
        ]

    def draw_pillow_start(self, drawing_area, pil_img):
        drawer = Drawer(self.core, drawing_area, pil_img)
        self.active_drawers[drawing_area] = drawer
        return drawer

    def draw_pillow_stop(self, drawing_area):
        drawer = self.active_drawers.pop(drawing_area)
        drawer.stop()
        return drawer


if __name__ == "__main__":
    import sys

    import PIL
    import PIL.Image
    import gi
    gi.require_version('Gtk', '3.0')
    from gi.repository import Gtk

    img = PIL.Image.open(sys.argv[1])

    core = openpaperwork_core.Core()
    core.load("openpaperwork_core.mainloop.asyncio")
    core.load("openpaperwork_gtk.fs.gio")
    core.load("openpaperwork_core.thread.simple")
    core.load("openpaperwork_core.work_queue.default")
    core.load("paperwork_backend.cairo.pillow")
    core.load("openpaperwork_core.pillow.img")
    core._load_module("test", sys.modules[__name__])
    core.init()

    window = Gtk.Window()
    window.set_size_request(600, 600)

    drawing_area = Gtk.DrawingArea()
    window.add(drawing_area)

    core.call_success("draw_pillow_start", drawing_area, img)

    window.show_all()

    Gtk.main()



import logging

from .. import PluginBase


LOGGER = logging.getLogger(__name__)


class Plugin(PluginBase):
    PRIORITY = 1000

    def config_load_plugins(self, plugin_list_name, default_plugins=[]):
        old_default = self.core.call_success(
            "config_backend_get", "default_plugins", plugin_list_name, None
        )
        if old_default is None:
            # no previously known list of default plugins. Weird but not much
            # we can do about it.
            self.core.call_all(
                "config_backend_put",
                "default_plugins", plugin_list_name, default_plugins
            )
            self.core.call_all("config_backend_save")
            return

        old_default = sorted(old_default)
        default_plugins = sorted(default_plugins)

        if old_default == default_plugins:
            # default list hasn't changed. So we assume the custom list
            # is still fine.
            return

        LOGGER.warning(
            "Default plugin list has changed."
            " Reseting plugin list to its new default."
        )
        LOGGER.warning("Old default list: %s", old_default)
        LOGGER.warning("New default list: %s", default_plugins)

        self.core.call_all("config_backend_reset_plugins", plugin_list_name)
        self.core.call_all(
            "config_backend_put",
            "default_plugins", plugin_list_name, default_plugins
        )
        self.core.call_all("config_backend_save")
